# coding: utf-8


import os

import h5py
import numpy
import pytest
from silx.io.url import DataUrl

from tomoscan.esrf.identifier.hdf5Identifier import HDF5VolumeIdentifier
from tomoscan.esrf.scan.mock import MockNXtomo as _MockNXtomo
from tomoscan.esrf.volume.hdf5volume import HDF5Volume, get_default_data_path_for_volume
from tomoscan.esrf.volume.mock import create_volume
from tomoscan.identifier import VolumeIdentifier

_data = create_volume(frame_dims=(20, 20), z_size=5)


_metadata = {
    "nabu_config": {
        "dataset": {
            "location": "toto.hdf5",
            "entry": "entry0000",
        },
        "phase": {
            "method": "None",
            "padding_type": "edge",
        },
    },
    "processing_option": {
        "build_sino": {
            "axis_correction": "None",
            "enable_halftomo": True,
        },
        "flatfield": {
            "binning": "11",
            "do_flat_distortion": False,
        },
    },
}


def test_hdf5volume_file_path(tmp_path):
    """test creation of an hdf5volume by providing url to the constructor"""
    test_dir = tmp_path / "test_volume_url"
    os.makedirs(test_dir)

    volume_file = os.path.join(test_dir, "volume.hdf5")
    assert not os.path.exists(volume_file)

    volume = HDF5Volume(
        file_path=volume_file,
        data_path="/entry0000/reconstruction",
        data=_data,
        metadata=_metadata,
    )
    assert volume.get_min_max_values() == (_data.min(), _data.max())

    url = DataUrl(file_path=volume_file, data_path="/entry0000/reconstruction")
    assert volume.url.file_path() == url.file_path()
    assert volume.url.data_path() == url.data_path()
    assert volume.data_url is not None
    assert volume.metadata_url is not None
    volume_identifier = volume.get_identifier()
    assert isinstance(volume_identifier, VolumeIdentifier)

    # insure data_url and metadata_url cannot be write
    with pytest.raises(AttributeError):
        volume.data_url = None

    with pytest.raises(AttributeError):
        volume.metadata_url = None

    volume.save()
    assert os.path.exists(volume_file)

    volume_loaded = HDF5Volume(
        file_path=volume_file, data_path="/entry0000/reconstruction"
    )
    assert volume_loaded._data is None
    assert volume_loaded._metadata is None
    # check loading function won't save loaded data in cache
    data_no_cache = volume_loaded.load_data(volume_loaded.data_url, store=False)
    metadata_no_cache = volume_loaded.load_metadata(
        volume_loaded.metadata_url, store=False
    )
    assert data_no_cache is not None
    assert metadata_no_cache is not None
    assert volume_loaded._data is None
    assert volume_loaded._metadata is None
    numpy.testing.assert_array_almost_equal(_data, data_no_cache)
    assert _metadata == metadata_no_cache
    # check data and metadata properties will load and set data in cache
    volume_loaded.load()
    data_cache = volume_loaded.data
    assert data_cache is not None
    metadata_cache = volume_loaded.metadata
    assert metadata_cache is not None
    numpy.testing.assert_array_almost_equal(_data, data_cache)
    assert _metadata == metadata_cache
    volume_loaded.data = None
    volume_loaded.metadata = None
    with pytest.raises(ValueError):
        volume_loaded.save_data()

    with pytest.raises(ValueError):
        volume_loaded.save_metadata()

    with pytest.raises(TypeError):
        volume_loaded.data = "toto"

    volume_loaded.data = create_volume(frame_dims=(20, 20), z_size=2)
    with pytest.raises(ValueError):
        volume_loaded.data = create_volume(frame_dims=(20, 20), z_size=2)[0]

    other_hdf5_file = os.path.join(test_dir, "other_hdf5_file.h5")
    new_volume_url = DataUrl(file_path=other_hdf5_file, data_path="data", scheme="silx")
    volume_loaded.save_data(url=new_volume_url)
    with pytest.raises(OSError):
        volume_loaded.save_data(
            url=DataUrl(file_path=other_hdf5_file, data_path="data", scheme="silx")
        )
    volume_loaded.overwrite = True
    volume_loaded.save_data(url=DataUrl(file_path=other_hdf5_file, data_path="data"))

    assert isinstance(volume_loaded.data, numpy.ndarray)
    assert isinstance(volume_loaded.load_data(url=new_volume_url), numpy.ndarray)
    numpy.testing.assert_array_almost_equal(
        volume_loaded.load_data(url=new_volume_url), volume_loaded.data
    )

    new_metadata_url = DataUrl(
        file_path=other_hdf5_file, data_path="metadata", scheme="silx"
    )
    with pytest.raises(ValueError):
        volume_loaded.save_metadata(url=new_metadata_url)

    volume_loaded.metadata = {"meta": False}
    volume_loaded.save_metadata(url=new_metadata_url)
    assert volume_loaded.load_metadata(url=new_metadata_url) == {"meta": False}

    # get the volume from the identifier
    volume_from_id = HDF5Volume.from_identifier(identifier=volume_identifier)
    assert volume_from_id.data is None
    assert volume_from_id.metadata is None

    # get the volume from the identifier str
    volume_from_str = HDF5Volume.from_identifier(
        identifier=HDF5VolumeIdentifier.from_str(volume_identifier.to_str())
    )
    assert volume_from_str.data is None
    assert volume_from_str.metadata is None

    # test browsing frames
    for l_frame, o_frame in zip(volume.browse_slices(), _data):
        numpy.testing.assert_allclose(l_frame, o_frame)

    # test hash
    hash(volume_from_str)

    # test get_min_max values
    assert volume_from_str.get_min_max() == (_data.min(), _data.max())
    volume_from_str.load_data()
    assert volume_from_str.get_min_max() == (_data.min(), _data.max())
    assert volume.data_extension == "hdf5"
    assert volume.metadata_extension == "hdf5"


def test_hdf5volume_data_url_and_metadata_url(tmp_path):
    """test creation of an hdf5volume by providing a data url and a metadata url"""
    test_dir = tmp_path / "test_volume_data_url_metadata_url"
    os.makedirs(test_dir)

    hdf5_data_file = os.path.join(test_dir, "my_data_file.hdf5")
    hdf5_metadata_file = os.path.join(test_dir, "my_metadata_file.hdf5")
    data_url = DataUrl(
        file_path=hdf5_data_file,
        data_path="data",
        scheme="silx",
    )
    metadata_url = DataUrl(
        file_path=hdf5_metadata_file,
        data_path="metadata",
        scheme="silx",
    )
    volume = HDF5Volume(
        data_url=data_url, metadata_url=metadata_url, data=_data, metadata=_metadata
    )

    with pytest.raises(ValueError):
        assert volume.get_identifier()

    assert tuple(volume.browse_data_files()) == tuple()
    assert tuple(volume.browse_data_urls()) == tuple()
    assert tuple(volume.browse_metadata_files()) == tuple()
    assert not os.path.exists(hdf5_data_file)
    volume.save_data()
    assert os.path.exists(hdf5_data_file)
    assert not os.path.exists(hdf5_metadata_file)
    volume.save_metadata()
    assert os.path.exists(hdf5_metadata_file)

    volume_loaded = HDF5Volume(data_url=data_url, metadata_url=metadata_url)
    assert volume_loaded.metadata is None
    volume_loaded.load_metadata()
    assert volume_loaded.metadata is not None
    assert volume_loaded.data is None

    # test get_slice
    numpy.testing.assert_array_equal(
        volume_loaded.get_slice(index=2, axis=0), _data[2, :, :]
    )
    numpy.testing.assert_array_equal(
        volume_loaded.get_slice(index=2, axis=1), _data[:, 2, :]
    )
    numpy.testing.assert_array_equal(
        volume_loaded.get_slice(index=2, axis=2), _data[:, :, 2]
    )

    # test old API (deprecated)
    numpy.testing.assert_array_equal(
        volume_loaded.get_slice(index=2, axis=0), volume_loaded.get_slice(xy=2)
    )
    numpy.testing.assert_array_equal(
        volume_loaded.get_slice(index=2, axis=1), volume_loaded.get_slice(xz=2)
    )
    numpy.testing.assert_array_equal(
        volume_loaded.get_slice(index=2, axis=2), volume_loaded.get_slice(yz=2)
    )

    volume_loaded.load_data()
    assert volume_loaded.data is not None

    # test get_slice
    numpy.testing.assert_array_equal(
        volume_loaded.get_slice(index=2, axis=0), _data[2, :, :]
    )
    numpy.testing.assert_array_equal(
        volume_loaded.get_slice(index=2, axis=1), _data[:, 2, :]
    )
    numpy.testing.assert_array_equal(
        volume_loaded.get_slice(index=2, axis=2), _data[:, :, 2]
    )

    numpy.testing.assert_array_equal(
        volume_loaded.data,
        _data,
    )

    assert tuple(volume_loaded.browse_data_files()) == (hdf5_data_file,)
    assert tuple(volume_loaded.browse_data_urls()) == (volume_loaded.data_url,)
    assert tuple(volume_loaded.browse_metadata_files()) == (hdf5_metadata_file,)

    # test load_chunk
    chunks_to_test = [
        (slice(None, None), slice(0, 5), slice(None, None)),
        (slice(0, 4), slice(0, 15), slice(None, None)),
    ]
    for chunk in chunks_to_test:
        data_chunk = volume.load_chunk(chunk)
        assert data_chunk.shape == _data[chunk].shape
        assert numpy.allclose(data_chunk, _data[chunk])


def test_write_hdf5_volume_with_parameter(tmp_path):
    test_dir = tmp_path / "test_volume_data_url_metadata_url"
    os.makedirs(test_dir)

    hdf5_data_file = os.path.join(test_dir, "my_data_file.hdf5")
    hdf5_metadata_file = os.path.join(test_dir, "my_metadata_file.hdf5")
    data_url = DataUrl(
        file_path=hdf5_data_file,
        data_path="data",
        scheme="silx",
    )
    metadata_url = DataUrl(
        file_path=hdf5_metadata_file,
        data_path="metadata",
        scheme="silx",
    )
    volume = HDF5Volume(
        data_url=data_url,
        metadata_url=metadata_url,
        data=_data,
        metadata=_metadata,
        overwrite=True,
    )

    # check chunks
    volume.save(chunks=True)
    with h5py.File(hdf5_data_file, mode="r") as h5f:
        assert h5f["data"].chunks is not None

    # check compression
    volume.save(compression="gzip", compression_opts=9)
    with h5py.File(hdf5_data_file, mode="r") as h5f:
        assert h5f["data"].compression == "gzip"


def test_data_file_saver_generator(tmp_path):
    """
    data_file_saver_generator (dumping frame by frame a volume to disk)
    """
    volume_file_path = str(tmp_path / "volume.h5")
    volume = HDF5Volume(file_path=volume_file_path, data_path="volume")
    for slice_, slice_saver in zip(
        _data,
        volume.data_file_saver_generator(
            n_frames=_data.shape[0], data_url=volume.data_url, overwrite=False
        ),
    ):
        slice_saver[:] = slice_
    assert volume.data is None
    numpy.testing.assert_array_equal(
        volume.load_data(),
        _data,
    )


def test_write_hdf5_with_virtual_layout_as_data(tmp_path):
    """
    insure HDF5Volume.data is safely handling virtual layout
    """
    test_dir = tmp_path / "test_volume_data_url_metadata_url"
    os.makedirs(test_dir)

    n_file = 5
    virtual_layout = h5py.VirtualLayout(shape=(n_file * 10, 100, 100), dtype=float)
    # create raw data
    for i_file in range(n_file):
        file_path = os.path.join(test_dir, f"file{i_file}.hdf5")
        data_path = f"path_to_dataset_{i_file}"
        with h5py.File(file_path, mode="w") as h5f:
            data = numpy.arange(0, 10 * 100 * 100).reshape(10, 100, 100)
            h5f[data_path] = data
            vs = h5py.VirtualSource(h5f[data_path])
            virtual_layout[i_file * 10 : (i_file + 1) * 10] = vs

    hdf5_data_file = os.path.join(test_dir, "my_data_file.hdf5")

    volume = HDF5Volume(
        file_path=hdf5_data_file,
        data_path="/",
        data=virtual_layout,
        metadata=_metadata,
        overwrite=True,
    )

    volume.save()
    with h5py.File(hdf5_data_file, mode="r") as h5f:
        assert h5f["results/data"].is_virtual
        numpy.testing.assert_array_equal(
            h5f["results/data"][:10],
            numpy.arange(0, 10 * 100 * 100).reshape(10, 100, 100),
        )
    volume.data = None
    assert volume.get_volume_shape() == (n_file * 10, 100, 100)


def test_utils(tmp_path):
    """test utils functions of the hdf5volume module"""

    test_dir = str(tmp_path / "test_volume_url")
    os.makedirs(test_dir)

    mock_scan = _MockNXtomo(
        scan_path=test_dir,
        dim=20,
        n_proj=12,
    )
    scan = mock_scan.scan

    assert isinstance(get_default_data_path_for_volume(scan), str)
    with pytest.raises(TypeError):
        get_default_data_path_for_volume("toto")


def test_example():
    """test static function 'example'"""
    assert isinstance(HDF5Volume.example_defined_from_str_identifier(), str)


@pytest.mark.parametrize("save_volume_to_file", (True, False))
def test_get_slices(tmp_path, save_volume_to_file: bool):
    """test the volume 'get_slices' function"""
    test_dir = tmp_path / "test_get_slices"
    test_dir.mkdir()

    volume_file = os.path.join(test_dir, "volume.hdf5")
    assert not os.path.exists(volume_file)
    data = numpy.linspace(0, 1000, num=1000, endpoint=False).reshape(10, 10, 10)

    volume = HDF5Volume(
        file_path=volume_file,
        data_path="/entry0000/reconstruction",
        data=data,
        metadata=_metadata,
    )
    if save_volume_to_file:
        volume.save_data()
        volume.clear_cache()

    slices_axis_0 = volume.get_slices(((0, 2), (0, 5)))
    numpy.testing.assert_array_almost_equal(
        slices_axis_0[(0, 2)],
        data[2],
    )

    slices_axis_1 = volume.get_slices(((1, 0), (1, 6)))
    numpy.testing.assert_array_almost_equal(
        slices_axis_1[(1, 6)],
        data[:, 6, :],
    )
    assert isinstance(slices_axis_1[(1, 6)], numpy.ndarray)

    slices_axis_2 = volume.get_slices(((2, 8), (2, 9)))
    numpy.testing.assert_array_almost_equal(
        slices_axis_2[(2, 9)],
        data[:, :, 9],
    )

    with pytest.raises(TypeError):
        volume.get_slices(
            "toto",
        )

    with pytest.raises(TypeError):
        volume.get_slices(
            (0, 10),
        )



# coding: utf-8
from __future__ import annotations

import os
import tempfile

import h5py
import numpy
import pytest
from silx.io.url import DataUrl

import tomoscan.esrf.scan.nxtomoscan
import tomoscan.normalization.normalization
from tomoscan.normalization.normalization import _DatasetInfos
from tomoscan.normalization.datasetscope import DatasetScope
from tomoscan.nexus.paths.nxtomo import nx_tomo_path_latest
from tomoscan.tests.utils import NXtomoMockContext

try:
    import scipy.interpolate  # noqa F401
except ImportError:
    has_scipy = False
else:
    has_scipy = True


@pytest.mark.parametrize("method", ["subtraction", "division"])
def test_normalization_scalar_normalization(method):
    """test scalar normalization"""
    with NXtomoMockContext(
        scan_path=os.path.join(tempfile.mkdtemp(), "scan_test"),
        n_proj=10,
        n_ini_proj=10,
    ) as scan:
        with pytest.raises(KeyError):
            scan.get_sinogram(line=2, norm_method=method)

        scan.get_sinogram(line=2, norm_method=method, value=12.2)


def test_normalize_chebyshev_2D():
    """Test checbychev 2D normalization"""
    with NXtomoMockContext(
        scan_path=os.path.join(tempfile.mkdtemp(), "scan_test"),
        n_proj=10,
        n_ini_proj=10,
    ) as scan:
        sinogram = scan.get_sinogram(line=2)
        tomoscan.normalization.normalize_chebyshev_2D(sinogram)
        sinogram_2 = scan.get_sinogram(line=2, norm_method="chebyshev")
        assert numpy.array_equal(sinogram, sinogram_2)


@pytest.mark.skipif(condition=not has_scipy, reason="scipy missing")
def test_normalize_lsqr_spline_2D():
    """test lsqr_spline_2D normalization"""
    with NXtomoMockContext(
        scan_path=os.path.join(tempfile.mkdtemp(), "scan_test"),
        n_proj=10,
        n_ini_proj=10,
    ) as scan:
        sinogram = scan.get_sinogram(line=2)
        tomoscan.normalization.normalize_lsqr_spline_2D(sinogram)
        sinogram_2 = scan.get_sinogram(line=2, norm_method="lsqr spline")
        assert numpy.array_equal(sinogram, sinogram_2)


def test_normalize_dataset():
    """Test extra information that can be provided relative to a dataset"""
    with NXtomoMockContext(
        scan_path=os.path.join(tempfile.mkdtemp(), "scan_test"),
        n_proj=10,
        n_ini_proj=10,
        dim=100,
        intensity_monitor=True,
    ) as scan:
        datasetinfo = _DatasetInfos()
        datasetinfo.file_path = scan.master_file
        datasetinfo.data_path = "/".join(
            [scan.entry, nx_tomo_path_latest.INTENSITY_MONITOR_PATH]
        )
        assert isinstance(datasetinfo.data_path, str)
        assert isinstance(datasetinfo.file_path, str)
        datasetinfo.scope = DatasetScope.GLOBAL
        assert isinstance(datasetinfo.scope, DatasetScope)
        scan.intensity_normalization.set_extra_infos(datasetinfo)


def test_normalize_roi():
    """Test extra information that can be provided relative to a roi"""
    with NXtomoMockContext(
        scan_path=os.path.join(tempfile.mkdtemp(), "scan_test"),
        n_proj=10,
        n_ini_proj=10,
        dim=100,
        intensity_monitor=True,
    ) as scan:
        roi_info = tomoscan.normalization.normalization._ROIInfo()
        scan.intensity_normalization.set_extra_infos(roi_info)
        scan.intensity_normalization.method = None
        scan.intensity_normalization.method = "lsqr spline"
        assert isinstance(
            scan.intensity_normalization.method, tomoscan.normalization.Method
        )
        str(scan.intensity_normalization)


@pytest.mark.parametrize("norm_method", ("subtraction", "division"))
@pytest.mark.parametrize("as_url", (True, False))
@pytest.mark.parametrize(
    "values",
    (
        0.5,
        numpy.arange(1, 11),
        numpy.arange(1, 101).reshape(10, 10),
        numpy.arange(1, 1001).reshape(10, 10, 10),
    ),
)
def test_get_sinogram(tmp_path, norm_method, as_url, values: float | numpy.ndarray):
    test_dir = tmp_path / "test1"
    test_dir.mkdir()
    params = {}
    if as_url:
        file_path = str(test_dir / "tmp_file.hdf5")
        with h5py.File(file_path, mode="w") as root:
            root["data"] = values
        params["dataset_url"] = DataUrl(
            file_path=file_path, data_path="data", scheme="silx"
        )
    else:
        params["value"] = values

    with NXtomoMockContext(
        scan_path=os.path.join(test_dir, "scan_test"),
        n_proj=10,
        n_ini_proj=10,
        dim=10,
    ) as scan:
        raw_sinogram = scan.get_sinogram(line=2)
        norm_sinogram = scan.get_sinogram(line=2, norm_method=norm_method, **params)
        assert isinstance(raw_sinogram, numpy.ndarray)
        assert isinstance(norm_sinogram, numpy.ndarray)
        if numpy.isscalar(values):
            if norm_method == "subtraction":
                numpy.testing.assert_almost_equal(norm_sinogram, raw_sinogram - values)
            elif norm_method == "division":
                numpy.testing.assert_almost_equal(norm_sinogram, raw_sinogram / values)
            else:
                raise ValueError
        elif values.ndim == 1:
            expected_sinogram = raw_sinogram.copy()
            for i_proj, proj_value in enumerate(values):
                if norm_method == "subtraction":
                    expected_sinogram[i_proj] = raw_sinogram[i_proj] - proj_value
                elif norm_method == "division":
                    expected_sinogram[i_proj] = raw_sinogram[i_proj] / proj_value
                else:
                    raise ValueError(norm_method)
            numpy.testing.assert_almost_equal(norm_sinogram, expected_sinogram)
        elif values.ndim in (2, 3):
            expected_sinogram = raw_sinogram.copy()
            for i_proj, proj_value in enumerate(values):
                if norm_method == "subtraction":
                    expected_sinogram[i_proj] = raw_sinogram[i_proj] - proj_value[2]
                elif norm_method == "division":
                    expected_sinogram[i_proj] = raw_sinogram[i_proj] / proj_value[2]
                else:
                    raise ValueError(norm_method)
            numpy.testing.assert_almost_equal(norm_sinogram, expected_sinogram)



"""module dedicated to esrf scans utils"""

from __future__ import annotations


import contextlib
import fnmatch
import logging
import os
import sys
import psutil
import warnings
import resource
import traceback

import fabio
import h5py
import numpy

from nxtomo.application.nxtomo import ImageKey

from silx.io.dictdump import h5todict
from silx.io.url import DataUrl
from silx.io.utils import get_data as silx_get_data
from silx.io.utils import h5py_read_dataset
from silx.utils.deprecation import deprecated
from silx.io.utils import open as hdf5_open

from tomoscan.utils.io import filter_esrf_mounting_points
from tomoscan.scanbase import ReducedFramesInfos, TomoScanBase

_logger = logging.getLogger(__name__)


__all__ = [
    "get_parameters_frm_par_or_info",
    "extract_urls_from_edf",
    "get_compacted_dataslices",
    "from_sequences_to_grps",
    "check_grp_is_valid",
    "grp_is_complete",
    "dataset_has_broken_vds",
    "check_possible_issue_with_rlimit",
    "get_datasets_linked_to_vds",
    "get_unique_files_linked",
    "get_files_from_pattern",
    "dump_info_file",
    "cwd_context",
    "get_data",
    "copy_h5_dict_darks_to",
    "copy_h5_dict_flats_to",
    "from_relative_reduced_frames_to_absolute",
    "from_absolute_reduced_frames_to_relative",
]


def get_parameters_frm_par_or_info(file_: str) -> dict:
    """
    Create a dictionary from the file with the information name as keys and
    their values as values

    :param file_: path to the file to parse
    :raises: ValueError when fail to parse some line.
    """
    assert os.path.exists(file_) and os.path.isfile(file_)
    ddict = {}
    f = open(file_, "r")
    lines = f.readlines()
    for line in lines:
        if "=" not in line:
            continue
        line_ = line.replace(" ", "")
        line_ = line_.rstrip("\n")
        # remove on the line comments
        if "#" in line_:
            line_ = line_.split("#")[0]
        if line_ == "":
            continue
        try:
            key, value = line_.split("=")
        except ValueError:
            raise ValueError(f"fail to extract information from {line_}")
        else:
            # try to cast the value on int, else float else don't
            try:
                value = int(value)
            except Exception:
                try:
                    value = float(value)
                except ValueError:
                    pass
            ddict[key.lower()] = value
    return ddict


def extract_urls_from_edf(
    file_: str, start_index: int | None, n_frames: int | None = None
) -> dict:
    r"""
    return one DataUrl for each frame contain in file\_

    :param file\_: path to the file to parse
    :param n_frames: Number of frames in each edf file (inferred if not told)
    :param start_index:
    """
    res = {}
    index = 0 if start_index is None else start_index
    if n_frames is None:
        with fabio.open(file_) as fabio_file:
            n_frames = fabio_file.nframes
    for i_frame in range(n_frames):
        res[index] = DataUrl(
            scheme="fabio",
            file_path=file_,
            data_slice=[
                i_frame,
            ],
        )
        index += 1
    return res


def get_compacted_dataslices(
    urls: dict,
    max_grp_size=None,
    return_merged_indices=False,
    return_url_set=False,
    subsampling=1,
):
    """
    Regroup urls to get the data more efficiently.
    Build a structure mapping files indices to information on
    how to load the data: `{indices_set: data_location}`
    where `data_location` contains contiguous indices.

    :param urls: Dictionary where the key is an integer and the value is
                      a silx `DataUrl`.
    :param max_grp_size: maximum size of url grps
    :param return_merged_indices: if True return the last merged indices.
                                       Deprecated
    :param return_url_set: return a set with url containing `urls` slices
                                and data path

    :return: Dictionary where the key is a list of indices, and the value is
             the corresponding `silx.io.url.DataUrl` with merged data_slice
    """

    def _convert_to_slice(idx):
        if numpy.isscalar(idx):
            return slice(idx, idx + 1)
        # otherwise, assume already slice object
        return idx

    def is_contiguous_slice(slice1, slice2):
        if numpy.isscalar(slice1):
            slice1 = slice(slice1, slice1 + 1)
        if numpy.isscalar(slice2):
            slice2 = slice(slice2, slice2 + 1)
        return slice2.start == slice1.stop

    def merge_slices(slice1, slice2):
        return slice(slice1.start, slice2.stop)

    if return_merged_indices is True:
        warnings.warn(
            "return_merged_indices is deprecated. It will be removed in version 0.8"
        )

    if max_grp_size is None:
        max_grp_size = sys.maxsize

    if subsampling is None:
        subsampling = 1

    sorted_files_indices = sorted(urls.keys())
    idx0 = sorted_files_indices[0]
    first_url = urls[idx0]
    merged_indices = [[idx0]]
    data_location = [
        [
            first_url.file_path(),
            first_url.data_path(),
            _convert_to_slice(first_url.data_slice()),
            first_url.scheme(),
        ]
    ]
    pos = 0
    grp_size = 0
    curr_fp, curr_dp, curr_slice, curr_scheme = data_location[pos]
    for idx in sorted_files_indices[1:]:
        url = urls[idx]
        next_slice = _convert_to_slice(url.data_slice())
        if (
            (grp_size <= max_grp_size)
            and (url.file_path() == curr_fp)
            and (url.data_path() == curr_dp)
            and is_contiguous_slice(curr_slice, next_slice)
            and (url.scheme() == curr_scheme)
        ):
            merged_indices[pos].append(idx)
            merged_slices = merge_slices(curr_slice, next_slice)
            data_location[pos][-2] = merged_slices
            curr_slice = merged_slices
            grp_size += 1
        else:  # "jump"
            pos += 1
            merged_indices.append([idx])
            data_location.append(
                [
                    url.file_path(),
                    url.data_path(),
                    _convert_to_slice(url.data_slice()),
                    url.scheme(),
                ]
            )
            curr_fp, curr_dp, curr_slice, curr_scheme = data_location[pos]
            grp_size = 0

    # Format result
    res = {}
    for ind, dl in zip(merged_indices, data_location):
        res.update(
            dict.fromkeys(
                ind,
                DataUrl(
                    file_path=dl[0], data_path=dl[1], data_slice=dl[2], scheme=dl[3]
                ),
            )
        )

    # Subsample
    if subsampling > 1:
        next_pos = 0
        for idx in sorted_files_indices:
            url = res[idx]
            ds = url.data_slice()
            res[idx] = DataUrl(
                file_path=url.file_path(),
                data_path=url.data_path(),
                data_slice=slice(next_pos + ds.start, ds.stop, subsampling),
            )
            n_imgs = ds.stop - (ds.start + next_pos)
            next_pos = abs(-n_imgs % subsampling)

    if return_url_set:
        url_set = {}
        for _, url in res.items():
            path = url.file_path(), url.data_path(), str(url.data_slice())
            url_set[path] = url

        if return_merged_indices:
            return res, merge_slices, url_set
        else:
            return res, url_set

    if return_merged_indices:
        return res, merged_slices
    else:
        return res


@deprecated(
    replacement="tomoscan.serie.from_sequences_to_series", since_version="0.8.0"
)
def from_sequences_to_grps(
    scans: list[TomoScanBase] | tuple[TomoScanBase, ...] | numpy.ndarray,
) -> tuple:
    from tomoscan.serie import sequences_to_series_from_sample_name

    return sequences_to_series_from_sample_name(scans)


@deprecated(replacement="tomoscan.serie.check_serie_is_valid", since_version="0.8.0")
def check_grp_is_valid(
    scans: list[TomoScanBase] | tuple[TomoScanBase, ...] | numpy.ndarray,
):
    from tomoscan.serie import check_series_is_consistent_frm_sample_name

    return check_series_is_consistent_frm_sample_name(scans)


@deprecated(replacement="tomoscan.serie.serie_is_complete", since_version="0.8.0")
def grp_is_complete(
    scans: list[TomoScanBase] | tuple[TomoScanBase, ...] | numpy.ndarray,
) -> bool:
    from tomoscan.serie import series_is_complete_from_group_size

    return series_is_complete_from_group_size(scans)


def __get_log_fct(log_level):
    if log_level is logging.WARNING:
        return _logger.warning
    elif log_level is logging.ERROR:
        return _logger.error
    elif log_level is logging.DEBUG:
        return _logger.debug
    elif log_level is logging.INFO:
        return _logger.info
    elif log_level is logging.CRITICAL:
        return _logger.critical
    else:
        raise ValueError("logging level unrecognized")


def dataset_has_broken_vds(
    url: DataUrl,
    raise_error=False,
    log_level=logging.WARNING,
    return_unique_files=False,
) -> bool:
    """
    check that the provided url is not a VDS with broken links.


    :param DataUrl url: url to the dataset to treat
    :param raise_error: if True and dataset not existing will raise an error
    :param return_unique_files: if True return unique files. As this step can be time consuming and reused it can sometimes be convenients
    """
    if not isinstance(url, DataUrl):
        raise TypeError(f"{url} is expected to be an instance of {DataUrl}")

    uniques_files = ()

    with hdf5_open(url.file_path()) as h5f:
        dataset = h5f.get(url.data_path(), None)
        if dataset is None:
            msg = f"no data found at {url.file_path()}://{url.data_path()}"
            if raise_error:
                raise ValueError(msg)
            else:
                __get_log_fct(log_level)(msg)
            if return_unique_files:
                return (True, None)
            else:
                return True

        if not dataset.is_virtual:
            if return_unique_files:
                return (False, (url.file_path(),))
            else:
                return False

        else:
            # free dataset in case it point to another file. Else won't free the object before calling get_unique_files_linked
            dataset = None
            h5f.close()
            uniques_files = get_unique_files_linked(url=url)
            missing_files = tuple(
                filter(
                    lambda file_: not os.path.exists(file_),
                    uniques_files,
                )
            )
            if len(missing_files) > 0:
                msg = f"dataset {url.file_path()} has broken virtual-dataset at {url.data_path()}. {missing_files} missing"

                __get_log_fct(log_level)(msg)
                if raise_error:
                    raise OSError(msg)
                if return_unique_files:
                    return (True, uniques_files)
                else:
                    return True

    if return_unique_files:
        return (False, uniques_files)
    else:
        return False


def check_possible_issue_with_rlimit(
    url: DataUrl | None,
    raise_error=False,
    log_level=logging.WARNING,
    delta_n_file=0,
    substract_current_open=True,
    unique_files: tuple | None = None,
) -> bool:
    """
    check that the provided url does not contain more external file than (ulimit - delta_ulimit).
    Else if this limit is reached we will probably met some troubles when reading data.
    Once this limit is reached - vds data will return 0 only - silently
    """
    if unique_files is None and url is None:
        raise ValueError("'unique_files' or 'url' should be provided")

    if unique_files is None:
        # first check if dataset is virtual, else skip test
        with hdf5_open(url.file_path()) as h5f:
            dataset = h5f.get(url.data_path(), None)
            if dataset is None:
                msg = f"no data found at {url.file_path()}://{url.data_path()}"
                if raise_error:
                    raise ValueError(msg)
                else:
                    __get_log_fct(log_level)(msg)
                return True

            if not dataset.is_virtual:
                return False

        unique_files = get_unique_files_linked(url=url)

    n_files = len(unique_files)
    if substract_current_open:
        current_process = psutil.Process()
        n_open_file_currently = len(current_process.open_files())
    else:
        n_open_file_currently = 0
    try:
        rlimit = resource.getrlimit(resource.RLIMIT_NOFILE)[0]
    except (ValueError, OSError):
        _logger.warning("Failed to check_possible_issue_with_rlimit")
    else:
        might_met_troubles = n_files > (rlimit - delta_n_file - n_open_file_currently)
        if might_met_troubles:
            msg = f"too much external files to open from {url.path()} - contains {n_files} external files. OS rlimit is set to {rlimit}"
            __get_log_fct(log_level)(msg)
            if raise_error:
                raise OSError(msg)


def get_datasets_linked_to_vds(url: DataUrl):
    """
    Return set([file-path, data_path]) linked to the provided url
    """
    if not isinstance(url, DataUrl):
        raise TypeError(f"{url} is expected to be an instance of {DataUrl}")
    start_file_path = url.file_path()
    start_dataset_path = url.data_path()
    start_dataset_slice = url.data_slice()
    if isinstance(start_dataset_slice, slice):
        start_dataset_slice = tuple(
            range(
                start_dataset_slice.start,
                start_dataset_slice.stop,
                start_dataset_slice.step or 1,
            )
        )
    virtual_dataset_to_treat = set()
    final_dataset = set()
    already_checked = set()

    # first datasets to be tested
    virtual_dataset_to_treat.add(
        (start_file_path, start_dataset_path, start_dataset_slice),
    )

    while len(virtual_dataset_to_treat) > 0:
        to_treat = list(virtual_dataset_to_treat)
        virtual_dataset_to_treat.clear()
        for file_path, dataset_path, dataset_slice in to_treat:
            if (file_path, dataset_path, dataset_slice) in already_checked:
                continue
            if os.path.exists(file_path):
                with hdf5_open(file_path) as h5f:
                    dataset = h5f[dataset_path]
                    if dataset.is_virtual:
                        for vs_info in dataset.virtual_sources():
                            min_frame_bound = vs_info.vspace.get_select_bounds()[0][0]
                            max_frame_bound = vs_info.vspace.get_select_bounds()[1][0]
                            if isinstance(dataset_slice, int):
                                if (
                                    not min_frame_bound
                                    <= dataset_slice
                                    <= max_frame_bound
                                ):
                                    continue
                            elif isinstance(dataset_slice, tuple):
                                if (
                                    min_frame_bound > dataset_slice[-1]
                                    or max_frame_bound < dataset_slice[0]
                                ):
                                    continue

                            with cwd_context():
                                os.chdir(os.path.dirname(file_path))
                                # Fixme: For now will look at the entire dataset of the n +1 file.
                                # if those can also contains virtual dataset and we want to handle
                                # the case a part of it is broken but not ours this should handle
                                # hyperslab
                                if vs_info.file_name != ".":
                                    # do not check self contained dataset
                                    virtual_dataset_to_treat.add(
                                        (
                                            os.path.abspath(vs_info.file_name),
                                            # avoid calling os.path.abspath if the dataset is in the same file. Otherwise mess up with paths
                                            str(vs_info.dset_name),
                                            None,
                                        )
                                    )
                    else:
                        final_dataset.add((file_path, dataset_path, dataset_slice))
                    dataset = None

            else:
                final_dataset.add((file_path, dataset_path, dataset_slice))
            already_checked.add((file_path, dataset_path, dataset_slice))
    return final_dataset


def get_unique_files_linked(url: DataUrl):
    """
    Return the list of unique files linked to the DataUrl without depth limitation
    """
    unique_files = set()
    datasets_linked = get_datasets_linked_to_vds(url=url)
    [unique_files.add(file_path) for (file_path, _, _) in datasets_linked]
    return unique_files


def get_files_from_pattern(file_pattern: str, pattern: str, research_dir: str) -> dict:
    """
    return: all files using a {pattern} to store the index. Key is the index and value is the file name
    """
    files_frm_patterm = {}
    if ("{" + pattern + "}") not in file_pattern:
        return files_frm_patterm
    if not isinstance(file_pattern, str):
        raise TypeError(f"file_pattern is expected to be str not {type(file_pattern)}")
    if not isinstance(pattern, str):
        raise TypeError(f"pattern is expected to be str not {type(pattern)}")
    if not isinstance(research_dir, str):
        raise TypeError(
            f"research_dir is expected to be a str not {type(research_dir)}"
        )
    if not os.path.exists(research_dir):
        raise FileNotFoundError(f"{research_dir} does not exists")
    # look for some index_zfill4
    file_path_fn = file_pattern.format(**{pattern: "*"})
    for file in os.listdir(research_dir):
        if fnmatch.fnmatch(file.lower(), file_path_fn.lower()):
            # try to deduce the index from pattern
            idx_start = file_pattern.find("{" + pattern + "}")
            idx_end = len(file_pattern.replace("{" + pattern + "}", "")) - idx_start
            idx_as_str = file[idx_start:-idx_end]
            if idx_as_str != "":  # handle case of an empty string
                try:
                    idx_as_int = int(idx_as_str)
                except ValueError:
                    _logger.warning("Could not determined")
                else:
                    files_frm_patterm[idx_as_int] = file

    return files_frm_patterm


def dump_info_file(
    file_path,
    tomo_n,
    scan_range,
    flat_n,
    flat_on,
    dark_n,
    dim_1,
    dim_2,
    col_beg,
    col_end,
    row_beg,
    row_end,
    pixel_size,
    distance,
    energy,
):
    # write the info file
    with open(file_path, "w") as info_file:
        info_file.write("TOMO_N=    " + str(tomo_n) + "\n")
        info_file.write("ScanRange= " + str(scan_range) + "\n")
        info_file.write("REF_N=     " + str(flat_n) + "\n")
        info_file.write("REF_ON=    " + str(flat_on) + "\n")
        info_file.write("DARK_N=    " + str(dark_n) + "\n")
        info_file.write("Dim_1=     " + str(dim_1) + "\n")
        info_file.write("Dim_2=     " + str(dim_2) + "\n")
        info_file.write("Col_beg=   " + str(col_beg) + "\n")
        info_file.write("Col_end=   " + str(col_end) + "\n")
        info_file.write("Row_beg=   " + str(row_beg) + "\n")
        info_file.write("Row_end=   " + str(row_end) + "\n")
        info_file.write("PixelSize= " + str(pixel_size) + "\n")
        info_file.write("Distance=  " + str(distance) + "\n")
        info_file.write("Energy=    " + str(energy) + "\n")


@contextlib.contextmanager
def cwd_context(new_cwd=None):
    try:
        curdir = os.getcwd()
    except Exception:
        traceback.print_stack(limit=3)
        _logger.error("Working directory has been delated. Will move to '~'")
        curdir = os.path.expanduser("~")
    try:
        if new_cwd is not None and os.path.isfile(new_cwd):
            new_cwd = os.path.dirname(new_cwd)
        if new_cwd not in (None, ""):
            os.chdir(new_cwd)
        yield
    finally:
        if curdir is not None:
            os.chdir(curdir)


def get_data(url: DataUrl):
    # update the current working dircetory for external dataset
    if url.file_path() is not None and h5py.is_hdf5(url.file_path()):
        # convert path to real path to insure it will be constant when changing current working directory
        file_path = os.path.abspath(url.file_path())
        with cwd_context(file_path):
            with hdf5_open(file_path) as h5f:
                if url.data_path() in h5f:
                    if url.data_slice() is None:
                        return h5py_read_dataset(h5f[url.data_path()])
                    else:
                        return h5py_read_dataset(
                            h5f[url.data_path()], index=url.data_slice()
                        )
    else:
        # for other file format don't need to do the same
        return silx_get_data(url)


def copy_h5_dict_darks_to(
    scan,
    darks_url: DataUrl,
    save: bool = False,
    raise_error_if_url_empty: bool = True,
    overwrite: bool = False,
):
    """
    :param TomwerScanBase scan: target to copy darks
    :param DataUrl darks_url: DataUrl to find darks to be copied
    :param save: should we save the darks to disk. If not will only be set on scan cache
    :param raise_error_if_url_empty: if the provided DataUrl lead to now data shoudl we raise an error (like file or dataset missing...)
    """
    from tomoscan.scanbase import TomoScanBase  # avoid cyclic import

    if not isinstance(scan, TomoScanBase):
        raise TypeError(
            f"scan is expected to be an instance of {TomoScanBase}. {type(scan)} provided"
        )
    if not isinstance(darks_url, DataUrl):
        raise TypeError(
            f"darks_url is expected to be an instance of {DataUrl}. {type(darks_url)} provided"
        )
    if darks_url.scheme() not in (None, "silx", "h5py"):
        raise ValueError("handled scheme are 'silx' and 'h5py'")
    try:
        with cwd_context(darks_url.file_path()):
            my_dict = h5todict(
                h5file=darks_url.file_path(),
                path=darks_url.data_path(),
            )
    except Exception as e:
        if raise_error_if_url_empty:
            raise e
        else:
            return
    data, metadata = ReducedFramesInfos.split_data_and_metadata(my_dict)
    # handle relative frame position if any
    data = from_relative_reduced_frames_to_absolute(reduced_frames=data, scan=scan)
    scan.set_reduced_darks(darks=data, darks_infos=metadata)
    if save:
        scan.save_reduced_darks(darks=data, darks_infos=metadata, overwrite=overwrite)


def copy_h5_dict_flats_to(
    scan,
    flats_url: DataUrl,
    save=False,
    raise_error_if_url_empty=True,
    overwrite: bool = False,
):
    """
    :param TomwerScanBase scan: target to copy darks
    :param DataUrl darks_url: DataUrl to find darks to be copied
    :param save: should we save the darks to disk. If not will only be set on scan cache
    :param raise_error_if_url_empty: if the provided DataUrl lead to now data shoudl we raise an error (like file or dataset missing...)
    """
    from tomoscan.scanbase import TomoScanBase  # avoid cyclic import

    if not isinstance(scan, TomoScanBase):
        raise TypeError(
            f"scan is expected to be an instance of {TomoScanBase}. {type(scan)} provided"
        )
    if not isinstance(flats_url, DataUrl):
        raise TypeError(
            f"flats_url is expected to be an instance of {DataUrl}. {type(flats_url)} provided"
        )
    if flats_url.scheme() not in (None, "silx", "h5py"):
        raise ValueError("handled scheme are 'silx' and 'h5py'")
    try:
        with cwd_context(flats_url.file_path()):
            my_dict = h5todict(
                h5file=flats_url.file_path(),
                path=flats_url.data_path(),
            )
    except Exception as e:
        if raise_error_if_url_empty:
            raise ValueError("DataUrl is not pointing to any data") from e
        else:
            return
    data, metadata = ReducedFramesInfos.split_data_and_metadata(my_dict)
    # handle relative frame position if any
    data = from_relative_reduced_frames_to_absolute(reduced_frames=data, scan=scan)
    scan.set_reduced_flats(flats=data, flats_infos=metadata)
    if save:
        scan.save_reduced_flats(flats=data, flats_infos=metadata, overwrite=overwrite)


def from_relative_reduced_frames_to_absolute(reduced_frames: dict, scan: TomoScanBase):
    if not isinstance(reduced_frames, dict):
        raise TypeError(
            f"reduced_frames is expected to be a dict, {type(reduced_frames)} provided"
        )
    if not isinstance(scan, TomoScanBase):
        raise TypeError(f"scan is expected to be a TomoScanBase, {type(scan)} provided")

    frame_n = len(scan.projections) + len(scan.darks) + len(scan.flats)

    def convert(index):
        if isinstance(index, str) and index.endswith("r"):
            return int(float(index[:-1]) * (frame_n - 1))
        else:
            return index

    return {convert(key): value for key, value in reduced_frames.items()}


def from_absolute_reduced_frames_to_relative(reduced_frames: dict, scan: TomoScanBase):
    if not isinstance(reduced_frames, dict):
        raise TypeError(
            f"reduced_frames is expected to be a dict, {type(reduced_frames)} provided"
        )
    if not isinstance(scan, TomoScanBase):
        raise TypeError(f"scan is expected to be a TomoScanBase, {type(scan)} provided")

    frame_n = len(scan.projections) + len(scan.darks) + len(scan.flats)

    def convert(index):
        if isinstance(index, str) and index.endswith("r"):
            return index
        else:
            return f"{int(index) / frame_n}r"

    return {convert(key): value for key, value in reduced_frames.items()}


def get_n_series(image_key_values: tuple | list, image_key_type: ImageKey) -> int:
    """
    Return the number of series of an image_key. Image key can be dark, flat, or projection.
    A series is defined as a contiguous elements in image_key_values

    :param image_key_values: list or tuple of image_keys to consider. Can be integers or tomoscan.esrf.scan.hdf5scan.ImageKey
    :param image_key_type: The kind of image key we want number of series for
    """
    image_key_type = ImageKey(image_key_type)
    if image_key_type is ImageKey.INVALID:
        raise ValueError(
            "we can't count Invalid image keys series because those are ignored from tomoscan"
        )
    image_key_values = [ImageKey(img_key) for img_key in image_key_values]

    # remove invalid frames
    image_key_values = numpy.array(
        image_key_values
    )  # for filtering invalid value a numpy array is requested
    image_key_values = image_key_values[image_key_values != ImageKey.INVALID]

    n_series = 0
    is_in_a_series = False
    for frame in image_key_values:
        if frame == image_key_type and not is_in_a_series:
            is_in_a_series = True
            n_series += 1
        elif frame != image_key_type:
            is_in_a_series = False
    return n_series


def get_series_slice(
    image_key_values: tuple[int], image_key_type: ImageKey, series_index: int
) -> slice | None:
    """
    return the DataUrl corresponding to the to the serie at `series_index` for the type of image_key_type.

    Examples: with the following image_key_values:
    (
        ImageKey.DARK_FIELD
        ImageKey.DARK_FIELD
        ImageKey.FLAT_FIELD
        ImageKey.FLAT_FIELD
        ImageKey.FLAT_FIELD
        ImageKey.PROJECTION
        ImageKey.PROJECTION
        ImageKey.PROJECTION
        ImageKey.PROJECTION
    )

    .. code-block:: python

        get_serie_url(image_key_values, ImageKey.FLAT_FIELD, serie_index=0) == slice(2, 5)
        get_serie_url(image_key_values, ImageKey.FLAT_FIELD, serie_index=1) == None
        get_serie_url(image_key_values, ImageKey.FLAT_FIELD, serie_index=-1) == slice(2, 5)
    """
    image_key_type = ImageKey(image_key_type)
    if series_index == -1:
        series_index = (
            get_n_series(
                image_key_values=image_key_values, image_key_type=image_key_type
            )
            - 1
        )
    n_series = -1
    start = None
    is_in_a_series = False
    for i_frame, frame in enumerate(image_key_values):
        frame = ImageKey(frame)
        if frame.value == image_key_type.value and not is_in_a_series:
            n_series += 1
            is_in_a_series = True
            if n_series == series_index:
                start = i_frame
        elif frame.value != image_key_type.value and is_in_a_series:
            is_in_a_series = False
            if n_series == series_index:
                return slice(start, i_frame, 1)
    if start is not None:
        return slice(start, len(image_key_values), 1)
    return None


def from_bliss_original_file_to_raw(bliss_original_file: str | None) -> str | None:
    """
    convert NXtomo 'bliss_original_files' to drac raw parameter (folder containing the raw)
    without some possible noise added by 'realpath' like '/mnt/multipath-shares' or '/gpfs/easy'
    """
    if bliss_original_file is None:
        return None

    bliss_original_file = filter_esrf_mounting_points(bliss_original_file)
    return os.path.dirname(bliss_original_file)



import os
import h5py
import numpy

from tomoscan.io import check_virtual_sources_exist, cast_to_default_types


def test_vds(tmp_path):
    h5_file_without_vds = os.path.join(tmp_path, "h5_file_without_vds.hdf5")

    with h5py.File(h5_file_without_vds, mode="w") as h5f:
        h5f["data"] = numpy.random.random((120, 120))
    assert check_virtual_sources_exist(h5_file_without_vds, "data")

    h5_file_with_vds = os.path.join(tmp_path, "h5_file_with_vds.hdf5")

    # create some dataset
    for i in range(4):
        filename = os.path.join(tmp_path, f"{i}.h5")
        with h5py.File(filename, mode="w") as h5f:
            h5f.create_dataset("data", (100,), "i4", numpy.arange(100))

    layout = h5py.VirtualLayout(shape=(4, 100), dtype="i4")
    for i in range(4):
        filename = os.path.join(tmp_path, f"{i}.h5")
        layout[i] = h5py.VirtualSource(filename, "data", shape=(100,))

    with h5py.File(h5_file_with_vds, mode="w") as h5f:
        # create the virtual dataset
        h5f.create_virtual_dataset("data", layout, fillvalue=-5)
    assert check_virtual_sources_exist(h5_file_with_vds, "data")


def test_cast_to_default_types():
    """test 'cast_to_default_types' function"""
    assert cast_to_default_types(
        {
            "a": numpy.float64(25.6),
            "b": {
                "c": numpy.zeros(5, dtype=numpy.int32),
                "d": numpy.ones(4, dtype=numpy.uint64),
            },
            "e": (numpy.float32(45.5), numpy.float32(5.5)),
            "f": numpy.int16(5),
        }
    ) == {
        "a": 25.6,
        "b": {
            "c": [0] * 5,
            "d": [1] * 4,
        },
        "e": [45.5, 5.5],
        "f": 5,
    }



# coding: utf-8

import h5py
import collections
import logging
import os
import shutil
import tempfile
import unittest
from glob import glob

import fabio.edfimage
import numpy
import pytest
import silx.io.utils
from silx.io.url import DataUrl
from silx.io.utils import get_data

from tomoscan.esrf.edfscan import EDFTomoScan
from tomoscan.esrf.mock import MockEDF
from tomoscan.factory import Factory
from tomoscan.scanbase import ReducedFramesInfos, TomoScanBase
from tomoscan.tests.datasets import GitlabDataset

logging.disable(logging.INFO)


class TestTomoBaseHashable(unittest.TestCase):
    """Make sure EDFTomoScan is hashable"""

    def setUp(self):
        self._folder = tempfile.mkdtemp()
        MockEDF.fastMockAcquisition(self._folder, n_radio=100)

    def tearDown(self):
        shutil.rmtree(self._folder)

    def test_is_hashable(self):
        tomo_base = TomoScanBase(scan=self._folder, type_="toto")
        self.assertTrue(isinstance(tomo_base, collections.abc.Hashable))
        tomo_scan = EDFTomoScan(self._folder)
        self.assertTrue(isinstance(tomo_scan, collections.abc.Hashable))


class TestScanFactory(unittest.TestCase):
    """Make sure the Scan factory is correctly working. Able to detect the valid
    scan type for a given file / directory
    """

    def test_no_scan(self):
        scan_dir = tempfile.mkdtemp()
        with self.assertRaises(ValueError):
            Factory.create_scan_object(scan_dir)

    def test_scan_edf(self):
        scan_dir = GitlabDataset.get_dataset("edf_datasets/test10")
        scan = Factory.create_scan_object(scan_dir)
        self.assertTrue(isinstance(scan, EDFTomoScan))
        assert scan.x_translation is None
        assert scan.y_translation is None
        assert scan.z_translation is None
        self.assertEqual(scan.tomo_n, 20)
        scan.clear_cache()
        scan.update()
        self.assertFalse(scan.is_abort())
        self.assertEqual(len(scan.projections), 20)
        self.assertEqual(len(scan.alignment_projections), 13)
        self.assertEqual(scan.scan_range, 180)
        self.assertEqual(scan.field_of_view, None)
        self.assertEqual(scan.x_rotation_axis_pixel_position, None)
        assert scan.dim_1 == 512
        assert scan.dim_2 == 512
        assert scan.ff_interval == 20
        assert len(scan.alignment_projections) > 0
        assert scan.instrument_name is None
        # this information is not provided for edf
        assert numpy.isclose(scan.sample_detector_distance, 0.008)
        assert numpy.isclose(scan.energy, 19.0)
        assert numpy.isclose(scan.pixel_size, 3.02e-06)
        assert numpy.isclose(scan.scan_range, 180)


class TestDarksFlats(unittest.TestCase):
    """unit test for the FTSerieReconstruction functions"""

    def setUp(self):
        def saveData(data, _file, folder):
            file_desc = fabio.edfimage.EdfImage(data=data)
            file_desc.write(os.path.join(folder, _file))

        unittest.TestCase.setUp(self)
        self.folderTwoFlats = tempfile.mkdtemp()
        self.folderOneFlat = tempfile.mkdtemp()

        self.darkData = numpy.arange(100).reshape(10, 10)
        self.flatField_0 = numpy.arange(start=10, stop=110).reshape(10, 10)
        self.flatField_600 = numpy.arange(start=-10, stop=90).reshape(10, 10)
        self.flatField = numpy.arange(start=0, stop=100).reshape(10, 10)

        # save configuration one
        saveData(self.darkData, "darkHST.edf", self.folderTwoFlats)
        saveData(self.flatField_0, "refHST_0000.edf", self.folderTwoFlats)
        saveData(self.flatField_600, "refHST_0600.edf", self.folderTwoFlats)

        # save configuration two
        saveData(self.darkData, "dark.edf", self.folderOneFlat)
        saveData(self.flatField, "refHST.edf", self.folderOneFlat)

        self.acquiOneFlat = EDFTomoScan(self.folderOneFlat)
        self.acquiTwoFlats = EDFTomoScan(self.folderTwoFlats)

    def tearDown(self):
        for f in (self.folderOneFlat, self.folderTwoFlats):
            shutil.rmtree(f)
        unittest.TestCase.tearDown(self)

    def testDarks(self):
        self.assertTrue(isinstance(self.acquiOneFlat.darks, dict))
        self.assertEqual(len(self.acquiOneFlat.darks), 1)
        self.assertTrue(isinstance(self.acquiTwoFlats.darks, dict))
        self.assertEqual(len(self.acquiTwoFlats.darks), 1)

        self.assertTrue(
            numpy.array_equal(
                silx.io.utils.get_data(self.acquiOneFlat.darks[0]), self.darkData
            )
        )
        self.assertTrue(
            numpy.array_equal(
                silx.io.utils.get_data(self.acquiTwoFlats.darks[0]), self.darkData
            )
        )

    def testFlats(self):
        """test that refHST flats are correctly found"""
        # check one flat file with two ref
        self.assertEqual(len(self.acquiOneFlat.flats), 1)
        self.assertTrue(isinstance(self.acquiOneFlat.flats[0], silx.io.url.DataUrl))

        data = silx.io.utils.get_data(self.acquiOneFlat.flats[0])
        numpy.array_equal(data, self.flatField)

        # check two flat files
        self.assertTrue(
            numpy.array_equal(
                silx.io.utils.get_data(self.acquiTwoFlats.flats[600]),
                self.flatField_600,
            )
        )
        self.assertTrue(
            numpy.array_equal(
                silx.io.utils.get_data(self.acquiTwoFlats.flats[0]), self.flatField_0
            )
        )
        self.assertTrue(12 not in self.acquiTwoFlats.flats)
        self.acquiTwoFlats.clear_frames_cache()
        os.remove(os.path.join(self.folderTwoFlats, "refHST_0000.edf"))
        os.remove(os.path.join(self.folderTwoFlats, "refHST_0600.edf"))
        self.assertEqual(self.acquiTwoFlats.flats, {})


class TestOriDarksFlats(unittest.TestCase):
    """
    Test that original ref* and dark* before creation of refHST  and darkHST
    are correctly found
    """

    def setUp(self) -> None:
        unittest.TestCase.setUp(self)
        self.pixel_size = 0.05
        self.det_width = 10
        self.det_height = 10

        self.folder = tempfile.mkdtemp()
        # create 2 series of ref with 20 frames each and 2 refHST file
        for ref_index in ("0000", "1500"):
            # ref files
            for ref_sub_index in range(20):
                file_desc = fabio.edfimage.EdfImage(
                    data=numpy.random.random((self.det_height, self.det_width))
                )
                file_name = "_".join(
                    (
                        "ref",
                        str(ref_sub_index).zfill(4),
                        str(ref_index).zfill(4) + ".edf",
                    )
                )
                file_name = os.path.join(self.folder, file_name)
                file_desc.write(file_name)
            # simulate two refHST generated by tomwer
            file_desc = fabio.edfimage.EdfImage(
                data=numpy.random.random((self.det_height, self.det_width))
            )
            file_name = "refHST" + str(ref_index).zfill(4) + ".edf"
            file_name = os.path.join(self.folder, file_name)
            file_desc.write(file_name)

        # create one dark with 10 frames
        file_desc = fabio.edfimage.EdfImage(
            data=numpy.random.random((self.det_height, self.det_width))
        )
        for frame in range(9):
            file_desc.append_frame(
                data=numpy.random.random((self.det_height, self.det_width))
            )
        self._dark_file_name = os.path.join(self.folder, "darkend0000.edf")
        file_desc.write(self._dark_file_name)

        self.scan = EDFTomoScan(self.folder)
        # create a .info for mocking the acquisition
        self.write_metadata(n_radio=0, scan_range=360, flat_n=20, dark_n=1)
        assert self.scan.dark_n == 1
        assert self.scan.flat_n == 20
        assert numpy.isclose(self.scan.pixel_size * 10e5, self.pixel_size)

    def get_info_file(self):
        return os.path.join(self.scan.path, os.path.basename(self.scan.path) + ".info")

    def write_metadata(self, n_radio, scan_range, flat_n, dark_n):
        info_file = self.get_info_file()
        if not os.path.exists(info_file):
            # write the info file
            with open(self.get_info_file(), "w") as info_file:
                info_file.write("TOMO_N=    " + str(n_radio) + "\n")
                info_file.write("ScanRange= " + str(scan_range) + "\n")
                info_file.write("REF_N=     " + str(flat_n) + "\n")
                info_file.write("REF_ON=    " + str(n_radio) + "\n")
                info_file.write("DARK_N=    " + str(dark_n) + "\n")
                info_file.write("Dim_1=     " + str(self.det_width) + "\n")
                info_file.write("Dim_2=     " + str(self.det_height) + "\n")
                info_file.write("Col_beg=    0" + "\n")
                info_file.write("Col_end=   " + str(self.det_width) + "\n")
                info_file.write("Row_beg=    0" + "\n")
                info_file.write("Row_end=    " + str(self.det_height) + "\n")
                info_file.write("PixelSize=  " + str(self.pixel_size) + "\n")

    def tearDown(self) -> None:
        shutil.rmtree(self.folder)
        unittest.TestCase.tearDown(self)

    def testFlatsOri(self):
        flat_urls = self.scan.get_flats_url(prefix="ref", scan_path=self.scan.path)
        self.assertEqual(len(flat_urls), 42)
        # checks some random files
        file_paths = [url.file_path() for url in flat_urls.values()]
        for ref_file in ("ref_0000_1500.edf", "ref_0004_0000.edf", "ref_0019_1500.edf"):
            ref_full_path = os.path.join(self.scan.path, ref_file)
            self.assertTrue(ref_full_path in file_paths)
        flat_urls = self.scan.get_flats_url(prefix="refHST", scan_path=self.scan.path)
        self.assertEqual(len(flat_urls), 2)

        flat_urls = self.scan.get_flats_url(
            prefix="ref", scan_path=self.scan.path, ignore=("HST",)
        )
        self.assertEqual(len(flat_urls), 40)

    def testDarksOri(self):
        darks = self.scan.get_darks_url(prefix="darkend", scan_path=self.scan.path)
        self.assertEqual(len(darks), 10)
        # check one random url
        url = DataUrl(
            file_path=self._dark_file_name,
            data_slice=[
                4,
            ],
            scheme="fabio",
        )
        self.assertTrue(url in darks.values())


class TestProjections(unittest.TestCase):
    """Test that the"""

    def setUp(self) -> None:
        self.folder = tempfile.mkdtemp()

    def tearDown(self) -> None:
        shutil.rmtree(self.folder)

    def testProjectionNoExtra(self):
        mock = MockEDF(
            scan_path=self.folder, n_radio=10, n_ini_radio=10, n_extra_radio=0
        )
        mock.end_acquisition()
        scan = EDFTomoScan(scan=self.folder)
        self.assertEqual(len(scan.projections), 10)

    def testProjectionUpdate(self):
        mock = MockEDF(scan_path=self.folder, n_radio=10, n_ini_radio=3)
        scan = EDFTomoScan(scan=self.folder)
        self.assertEqual(len(scan.projections), 3)
        mock.add_radio()
        self.assertEqual(len(scan.projections), 3)
        scan.update()
        self.assertEqual(len(scan.projections), 4)

        index_0 = list(scan.projections.keys())[0]
        self.assertTrue(isinstance(scan.projections[index_0], silx.io.url.DataUrl))

    def testProjectionWithExtraRadio(self):
        mock = MockEDF(
            scan_path=self.folder,
            n_radio=11,
            n_ini_radio=11,
            n_extra_radio=2,
            scan_range=180,
        )
        mock.end_acquisition()
        scan = EDFTomoScan(scan=self.folder)
        self.assertEqual(len(scan.projections), 11)
        self.assertEqual(len(scan.alignment_projections), 2)
        proj_angle_dict = scan.get_proj_angle_url()
        self.assertEqual(len(proj_angle_dict), 11 + 2)
        self.assertTrue(90 in proj_angle_dict)
        self.assertTrue(180 in proj_angle_dict)
        self.assertTrue("90(1)" in proj_angle_dict)
        self.assertTrue("0(1)" in proj_angle_dict)
        self.assertTrue(360 not in proj_angle_dict)


class TestFlatFieldCorrection(unittest.TestCase):
    """Test the flat field correction"""

    def setUp(self) -> None:
        self.folder = tempfile.mkdtemp()
        mock = MockEDF(
            scan_path=self.folder, n_radio=30, n_ini_radio=30, n_extra_radio=0, dim=20
        )
        mock.end_acquisition()
        self.scan = EDFTomoScan(scan=self.folder)

    def tearDown(self) -> None:
        shutil.rmtree(self.folder)

    def testFlatAndDarksSet(self):
        """Test no error is log if `normed` dark and flat are provided"""
        self.scan.set_reduced_flats(
            {
                1: numpy.random.random(20 * 20).reshape((20, 20)),
                21: numpy.random.random(20 * 20).reshape((20, 20)),
            }
        )
        self.scan.set_reduced_darks({0: numpy.random.random(20 * 20).reshape((20, 20))})

        projs = []
        proj_indexes = []
        for proj_index, proj in self.scan.projections.items():
            projs.append(proj)
            proj_indexes.append(proj_index)
        normed_proj = self.scan.flat_field_correction(
            projs=projs, proj_indexes=proj_indexes
        )
        self.assertEqual(len(normed_proj), len(self.scan.projections))
        raw_data = get_data(projs[10])
        self.assertFalse(numpy.array_equal(raw_data, normed_proj[10]))

    def testNoFlatOrDarkSet(self):
        """Test an error is log if `normed` dark and flat aren't provided"""
        projs = []
        proj_indexes = []
        for proj_index, proj in self.scan.projections.items():
            projs.append(proj)
            proj_indexes.append(proj_index)
        with self.assertLogs("tomoscan", level="ERROR"):
            normed_proj = self.scan.flat_field_correction(
                projs=projs, proj_indexes=proj_indexes
            )
        self.assertEqual(len(normed_proj), len(self.scan.projections))
        raw_data = get_data(projs[10])
        numpy.testing.assert_array_equal(raw_data, normed_proj[10])


class TestGetSinogram(unittest.TestCase):
    """Test the get_sinogram function"""

    def setUp(self) -> None:
        self.folder = tempfile.mkdtemp()
        mock = MockEDF(
            scan_path=self.folder, n_radio=30, n_ini_radio=30, n_extra_radio=0, dim=20
        )
        mock.end_acquisition()
        self.scan = EDFTomoScan(scan=self.folder)
        self.scan.set_reduced_flats(
            {
                21: numpy.random.random(20 * 20).reshape((20, 20)),
                1520: numpy.random.random(20 * 20).reshape((20, 20)),
            }
        )
        self.scan.set_reduced_darks({0: numpy.random.random(20 * 20).reshape((20, 20))})

    def tearDown(self) -> None:
        shutil.rmtree(self.folder)

    def testGetSinogram1(self):
        sinogram = self.scan.get_sinogram(line=12, subsampling=1)
        self.assertEqual(sinogram.shape, (30, 20))

    def testGetSinogram2(self):
        """Test if subsampling is negative"""
        with self.assertRaises(ValueError):
            self.scan.get_sinogram(line=0, subsampling=-1)

    def testGetSinogram3(self):
        sinogram = self.scan.get_sinogram(line=0, subsampling=3)
        self.assertEqual(sinogram.shape, (10, 20))

    def testGetSinogram4(self):
        """Test if line is not in the projection"""
        with self.assertRaises(ValueError):
            self.scan.get_sinogram(line=-1, subsampling=1)

    def testGetSinogram5(self):
        """Test if line is not in the projection"""
        with self.assertRaises(ValueError):
            self.scan.get_sinogram(line=35, subsampling=1)


class TestScanValidatorFindFiles(unittest.TestCase):
    """Function testing the getReconstructionsPaths function is correctly
    functioning"""

    DIM_MOCK_SCAN = 10

    N_RADIO = 20
    N_RECONS = 10
    N_PAG_RECONS = 5

    def setUp(self):
        # create scan folder
        self.path = tempfile.mkdtemp()
        MockEDF.mockScan(
            scanID=self.path,
            nRadio=self.N_RADIO,
            nRecons=self.N_RECONS,
            nPagRecons=self.N_PAG_RECONS,
            dim=self.DIM_MOCK_SCAN,
        )
        basename = os.path.basename(self.path)

        # add some random files
        for _file in ("45gfdgfg1.edf", "465slicetest1.edf", "slice_ab.edf"):
            with open(os.path.join(self.path, basename + _file), "w+") as ofile:
                ofile.write("test")

    def tearDown(self):
        if os.path.isdir(self.path):
            shutil.rmtree(self.path)


class TestRadioPath(unittest.TestCase):
    """Test static method getRadioPaths for EDFTomoScan"""

    def test(self):
        files = [
            "essai1_0008.edf",
            "essai1_0019.edf",
            "essai1_0030.edf",
            "essai1_0041.edf",
            "essai1_0052.edf",
            "essai1_0063.edf",
            "essai1_0074.edf",
            "essai1_0085.edf",
            "essai1_0096.edf",
            "essai1_.par",
            "refHST0100.edf",
            "darkend0000.edf",
            "essai1_0009.edf",
            "essai1_0020.edf",
            "essai1_0031.edf",
            "essai1_0042.edf",
            "essai1_0053.edf",
            "essai1_0064.edf",
            "essai1_0075.edf",
            "essai1_0086.edf",
            "essai1_0097.edf",
            "essai1_.rec",
            "essai1_0000.edf",
            "essai1_0010.edf",
            "essai1_0021.edf",
            "essai1_0032.edf",
            "essai1_0043.edf",
            "essai1_0054.edf",
            "essai1_0065.edf",
            "essai1_0076.edf",
            "essai1_0087.edf",
            "essai1_0098.edf",
            "essai1_slice_1023.edf",
            "essai1_0001.edf",
            "essai1_0011.edf",
            "essai1_0022.edf",
            "essai1_0033.edf",
            "essai1_0044.edf",
            "essai1_0055.edf",
            "essai1_0066.edf",
            "essai1_0077.edf",
            "essai1_0088.edf",
            "essai1_0099.edf",
            "essai1_slice.info",
            "essai1_0001.par",
            "essai1_0012.edf",
            "essai1_0023.edf",
            "essai1_0034.edf",
            "essai1_0045.edf",
            "essai1_0056.edf",
            "essai1_0067.edf",
            "essai1_0078.edf",
            "essai1_0089.edf",
            "essai1_0100.edf",
            "essai1_slice.par",
            "essai1_0002.edf",
            "essai1_0013.edf",
            "essai1_0024.edf",
            "essai1_0035.edf",
            "essai1_0046.edf",
            "essai1_0057.edf",
            "essai1_0068.edf",
            "essai1_0079.edf",
            "essai1_0090.edf",
            "essai1_0101.edf",
            "essai1_slice.xml",
            "essai1_0003.edf",
            "essai1_0014.edf",
            "essai1_0025.edf",
            "essai1_0036.edf",
            "essai1_0047.edf",
            "essai1_0058.edf",
            "essai1_0069.edf",
            "essai1_0080.edf",
            "essai1_0091.edf",
            "essai1_0102.edf",
            "essai1_.xml",
            "essai1_0004.edf",
            "essai1_0015.edf",
            "essai1_0026.edf",
            "essai1_0037.edf",
            "essai1_0048.edf",
            "essai1_0059.edf",
            "essai1_0070.edf",
            "essai1_0081.edf",
            "essai1_0092.edf",
            "essai1_0103.edf",
            "histogram_essai1_slice",
            "essai1_0005.edf",
            "essai1_0016.edf",
            "essai1_0027.edf",
            "essai1_0038.edf",
            "essai1_0049.edf",
            "essai1_0060.edf",
            "essai1_0071.edf",
            "essai1_0082.edf",
            "essai1_0093.edf",
            "essai1_0104.edf",
            "machinefile",
            "essai1_0006.edf",
            "essai1_0017.edf",
            "essai1_0028.edf",
            "essai1_0039.edf",
            "essai1_0050.edf",
            "essai1_0061.edf",
            "essai1_0072.edf",
            "essai1_0083.edf",
            "essai1_0094.edf",
            "essai1_.cfg",
            "pyhst_out.txt",
            "essai1_0007.edf",
            "essai1_0018.edf",
            "essai1_0029.edf",
            "essai1_0040.edf",
            "essai1_0051.edf",
            "essai1_0062.edf",
            "essai1_0073.edf",
            "essai1_0084.edf",
            "essai1_0095.edf",
        ]

        nbRadio = 0
        for f in files:
            nbRadio += EDFTomoScan.is_a_proj_path(f, "essai1_")
        self.assertTrue(nbRadio == 105)


class TestIgnoredProjections(unittest.TestCase):
    """Test the ignore_projections parameter"""

    def setUp(self) -> None:
        self.folder = tempfile.mkdtemp()
        self.mock = MockEDF(self.folder, 18, n_ini_radio=18, dim=10, scan_range=180)
        self.all_angles = numpy.linspace(0, 180, self.mock.n_radio, False)
        self.mock.end_acquisition()

    def tearDown(self) -> None:
        shutil.rmtree(self.folder)

    def testIgnoreProjectionsIndices(self):
        ignored_projs_indices = [5, 6, 7]
        scan = EDFTomoScan(scan=self.folder, ignore_projections=ignored_projs_indices)
        for idx in ignored_projs_indices:
            self.assertFalse(
                idx in scan.projections,
                "Projection index %d is supposed to be ignored" % idx,
            )

    def testIgnoreProjectionAngles(self):
        ignored_projs_angles = [40.0, 50.0, 60.0, 83.3333, 156.5]
        # In this case idx k <-> angle 10*k
        corresponding_projs_indices = [4, 5, 6, 8, 16]
        scan = EDFTomoScan(
            scan=self.folder,
            ignore_projections={"kind": "angles", "values": ignored_projs_angles},
        )
        for idx in scan.projections.keys():
            assert_func = (
                self.assertFalse
                if idx in corresponding_projs_indices
                else self.assertTrue
            )
            assert_func(idx in scan.projections, "Projection index %d" % idx)

    def testIgnoreProjectionAngularRange(self):
        ignored_projs_angular_range = [95.5, 120]
        # In this case idx k <-> angle 10*k
        corresponding_projs_indices = [10, 11, 12]
        scan = EDFTomoScan(
            scan=self.folder,
            ignore_projections={"kind": "range", "values": ignored_projs_angular_range},
        )
        for idx in scan.projections.keys():
            assert_func = (
                self.assertFalse
                if idx in corresponding_projs_indices
                else self.assertTrue
            )
            assert_func(
                idx in scan.projections,
                "Something off with projection index %d. The scan has projections %s, but should have the following excluded: %s"
                % (idx, sorted(scan.projections.keys()), corresponding_projs_indices),
            )


def test_EDFTomoScan_API():
    """some simple test of the EDFTomoScan API"""
    with tempfile.TemporaryDirectory() as folder:
        MockEDF.mockScan(
            scanID=folder,
            n_extra_radio=2,
            scan_range=180,
            dim=36,
        )

        scan = EDFTomoScan(folder)
        scan.clear_cache()
        for fct in (
            "get_flat_expected_location",
            "get_dark_expected_location",
            "get_projection_expected_location",
            "get_energy_expected_location",
            "get_sample_detector_distance_expected_location",
            "get_pixel_size_expected_location",
        ):
            assert isinstance(getattr(scan, fct)(), str)
        # check str conversion
        str(scan)


def test_get_relative_file(tmpdir):
    """Test that get_relative_file function is working correctly for EDFScan"""
    folder_path = os.path.join(tmpdir, "myscan")
    MockEDF.mockScan(
        scanID=folder_path,
        n_extra_radio=2,
        scan_range=180,
        dim=36,
    )

    scan = EDFTomoScan(folder_path)
    expected_f1 = os.path.join(folder_path, "myscan_tomwer_processes.h5")
    assert (
        scan.get_relative_file("tomwer_processes.h5", with_dataset_prefix=True)
        == expected_f1
    )

    expected_f2 = os.path.join(folder_path, "tomwer_processes.h5")
    assert (
        scan.get_relative_file("tomwer_processes.h5", with_dataset_prefix=False)
        == expected_f2
    )


def test_save_and_load_dark(tmp_path):
    test_dir = tmp_path / "test_dir"
    test_dir.mkdir()

    MockEDF.mockScan(
        scanID=str(test_dir),
        n_extra_radio=2,
        scan_range=180,
        dim=36,
    )

    scan = EDFTomoScan(test_dir)
    assert scan.load_reduced_darks() == {}
    assert scan.load_reduced_flats() == {}
    dark_frame = numpy.ones((100, 100))
    scan.save_reduced_darks({0: dark_frame})
    loaded_darks = scan.load_reduced_darks()
    assert 0 in loaded_darks
    numpy.testing.assert_array_equal(loaded_darks[0], dark_frame)
    assert scan.load_reduced_flats() == {}


def test_save_and_load_flats(tmp_path):
    test_dir = tmp_path / "test_dir"
    test_dir.mkdir()

    MockEDF.mockScan(
        scanID=str(test_dir),
        n_extra_radio=2,
        scan_range=180,
        dim=36,
        nRecons=0,
    )

    scan = EDFTomoScan(test_dir)
    assert scan.load_reduced_darks() == {}
    flat_frame_1 = numpy.ones((100, 100))
    flat_frame_1000 = numpy.zeros((100, 100))
    assert len(glob(os.path.join(test_dir, "*.hdf5"))) == 0

    scan.save_reduced_flats(
        {
            1: flat_frame_1,
            1000: flat_frame_1000,
        }
    )
    # check hdf5 as also been created
    assert tuple(glob(os.path.join(test_dir, "*.hdf5"))) == (
        os.path.join(test_dir, "test_dir_flats.hdf5"),
    )

    loaded_flats = scan.load_reduced_flats()
    assert 1 in loaded_flats
    assert 1000 in loaded_flats
    numpy.testing.assert_array_equal(loaded_flats[1], flat_frame_1)
    numpy.testing.assert_array_equal(loaded_flats[1000], flat_frame_1000)
    assert scan.load_reduced_darks() == {}


def test_save_dark_flat_reduced_several_urls(tmp_path):
    """test saving and loading dark and flat providing several urls"""
    test_dir = tmp_path / "test_dir"
    test_dir.mkdir()

    MockEDF.mockScan(
        scanID=str(test_dir),
        n_extra_radio=2,
        scan_range=180,
        dim=36,
    )
    scan = EDFTomoScan(test_dir)
    url_flats_edf = EDFTomoScan.REDUCED_FLATS_DATAURLS[0]
    url_flats_processes = DataUrl(
        file_path=scan.get_relative_file(
            file_name="my_processes.h5", with_dataset_prefix=False
        ),
        data_path="/{entry}/process_1/results/flats/{index}",
        scheme="hdf5",
    )
    url_darks_edf = EDFTomoScan.REDUCED_DARKS_DATAURLS[0]
    url_darks_processes = DataUrl(
        file_path=scan.get_relative_file(
            file_name="my_processes.h5", with_dataset_prefix=False
        ),
        data_path="/{entry}/process_1/results/darks/{index}",
        scheme="silx",
    )

    scan = EDFTomoScan(test_dir)
    assert scan.load_reduced_darks(return_info=False) == {}
    flat_frame_1 = numpy.ones((100, 100))
    flat_frame_1000 = numpy.ones((100, 100)) * 1.2
    dark_frame = numpy.zeros((100, 100))
    scan.save_reduced_flats(
        flats={
            1: flat_frame_1,
            1000: flat_frame_1000,
        },
        output_urls=(
            url_flats_edf,
            url_flats_processes,
        ),
    )
    scan.save_reduced_darks(
        darks={
            0: dark_frame,
        },
        output_urls=(
            url_darks_edf,
            url_darks_processes,
        ),
    )
    assert len(scan.load_reduced_flats(return_info=False)) == 2
    assert len(scan.load_reduced_darks(return_info=False)) == 1

    _, load_infos_darks = scan.load_reduced_darks(return_info=True)
    assert load_infos_darks == ReducedFramesInfos()  # no data saved
    _, load_infos_flats = scan.load_reduced_flats(return_info=True)
    assert load_infos_flats == ReducedFramesInfos()  # no data saved

    processes_file = os.path.join(test_dir, "my_processes.h5")
    assert os.path.exists(processes_file)
    with h5py.File(processes_file) as h5s:
        assert "/entry/process_1/results/darks" in h5s
        darks_grp = h5s["/entry/process_1/results/darks"]
        assert "0" in darks_grp
        numpy.testing.assert_array_equal(darks_grp["0"][()], dark_frame)
        assert "/entry/process_1/results/flats" in h5s
        flats_grp = h5s["/entry/process_1/results/flats"]
        assert "1" in flats_grp
        assert "1000" in flats_grp
        numpy.testing.assert_array_equal(flats_grp["1"], flat_frame_1)
        numpy.testing.assert_array_equal(flats_grp["1000"], flat_frame_1000)

    # check flats when provided as DataUrl as HDF5File
    loaded_reduced_darks_hdf5 = scan.load_reduced_darks(
        (url_darks_processes,), return_as_url=True
    )
    assert isinstance(loaded_reduced_darks_hdf5[0], DataUrl)
    assert loaded_reduced_darks_hdf5[0].file_path() == processes_file
    assert (
        loaded_reduced_darks_hdf5[0].data_path() == "/entry/process_1/results/darks/0"
    )
    assert loaded_reduced_darks_hdf5[0].scheme() == "silx"
    loaded_reduced_flats = scan.load_reduced_flats(
        (url_flats_processes,), return_as_url=True
    )
    assert len(loaded_reduced_flats) == 2
    assert len(loaded_reduced_darks_hdf5) == 1

    # check flats when provided as DataUrl as EDFFile
    loaded_reduced_darks_edf = scan.load_reduced_darks(return_as_url=True)
    assert isinstance(loaded_reduced_darks_edf[0], DataUrl)
    assert loaded_reduced_darks_edf[0].file_path() == os.path.join(
        scan.path, "test_dir_darks.hdf5"
    )
    assert loaded_reduced_darks_edf[0].data_path() == "entry/darks/0"
    assert loaded_reduced_darks_edf[0].scheme() == "silx"

    # test metadata
    flats_infos = ReducedFramesInfos()
    flats_infos.count_time = [1.0, 1.0]
    flats_infos.machine_current = [13.0, 13.0]

    scan.save_reduced_flats(
        flats={
            1: flat_frame_1,
            1000: flat_frame_1000,
        },
        flats_infos=flats_infos,
        overwrite=True,
    )
    darks_infos = ReducedFramesInfos()
    darks_infos.count_time = [
        2.0,
    ]
    darks_infos.machine_current = [13.1]
    scan.save_reduced_darks(
        darks={
            0: dark_frame,
        },
        darks_infos=darks_infos,
        overwrite=True,
    )
    _, load_infos_darks = scan.load_reduced_darks(return_info=True)
    assert load_infos_darks == darks_infos
    _, load_infos_flats = scan.load_reduced_flats(return_info=True)
    assert load_infos_flats == flats_infos


folder_test_input = (
    "myacquisition",
    "my_acquisition",
    "my_acquisition_",
    "_my_acquisition_",
)
folder_test_output = ("new_dir", "new_dir_", "newdir", "_new_dir")


@pytest.mark.parametrize("original_dir_name", folder_test_input)
@pytest.mark.parametrize("final_dir_name", folder_test_output)
def test_EDFTomoScan_provide_file_prefix(tmp_path, original_dir_name, final_dir_name):
    """Test EDFTomoScan works if the file_prefix is provided and if folder name is different that the file_prefix"""
    original_dir = tmp_path / original_dir_name
    original_dir.mkdir()
    new_dir = tmp_path / final_dir_name

    n_projection = 6

    MockEDF.mockScan(
        scanID=str(original_dir),
        n_extra_radio=2,
        scan_range=180,
        nRadio=n_projection,
        dim=36,
    )
    shutil.copytree(original_dir, new_dir)
    scan = EDFTomoScan(new_dir, dataset_basename=original_dir_name)
    assert len(scan.projections) == n_projection, "not all projection were found"


def test_EDFTomoScan_provide_scan_info(tmp_path):
    """Test EDFTomoScan works as expected if scan_info is provided and insure .info information is overwrite"""
    original_dir = tmp_path / "my_acquisition"
    original_dir.mkdir()

    n_projection = 6

    MockEDF.mockScan(
        scanID=str(original_dir),
        n_extra_radio=2,
        scan_range=180,
        nRadio=n_projection,
        dim=36,
    )
    scan = EDFTomoScan(
        str(original_dir),
    )
    assert scan.scan_range == 180

    scan = EDFTomoScan(str(original_dir), scan_info={"ScanRange": 360})
    assert scan.scan_range == 360



# coding: utf-8
"""module defining utils for an hdf5 volume"""

from __future__ import annotations

import logging
import os

import h5py
import numpy
from silx.io.dictdump import dicttonx, nxtodict
from silx.io.url import DataUrl
from silx.io.utils import open as open_hdf5

from tomoscan.utils.io import deprecated_warning
from tomoscan.esrf.identifier.hdf5Identifier import HDF5VolumeIdentifier
from tomoscan.scanbase import TomoScanBase
from tomoscan.utils import docstring
from tomoscan.utils.io import _deprecate_url_in_signature
from tomoscan.volumebase import VolumeBase, SliceTuple

_logger = logging.getLogger(__name__)

__all__ = ["HDF5Volume", "get_default_data_path_for_volume"]


class HDF5Volume(VolumeBase):
    """
    Volume where both data and metadata are store in a HDF5 file but at a different location.
    """

    DATA_DATASET_NAME = "results/data"
    METADATA_GROUP_NAME = "configuration"

    def __init__(
        self,
        file_path: str | None = None,
        data_path: str | None = None,
        data: numpy.ndarray | None = None,
        source_scan: TomoScanBase | None = None,
        metadata: dict | None = None,
        data_url: DataUrl | None = None,
        metadata_url: DataUrl | None = None,
        overwrite: bool = False,
    ) -> None:
        url = self._get_url_from_file_path_data_path(
            file_path=file_path, data_path=data_path
        )

        self._file_path = file_path
        self._data_path = data_path
        super().__init__(
            url=url,
            data=data,
            source_scan=source_scan,
            metadata=metadata,
            data_url=data_url,
            metadata_url=metadata_url,
            overwrite=overwrite,
        )

    @property
    def data_extension(self) -> None | str:
        if self.data_url is not None and self.data_url.file_path() is not None:
            _, extension = os.path.splitext(self.data_url.file_path())
            if extension:
                return extension[1:]

    @property
    def metadata_extension(self) -> None | str:
        if self.metadata_url is not None and self.metadata_url.file_path() is not None:
            _, extension = os.path.splitext(self.metadata_url.file_path())
            if extension:
                return extension[1:]

    @staticmethod
    def _get_url_from_file_path_data_path(
        file_path: str | None, data_path: str | None
    ) -> DataUrl | None:
        if file_path is not None and data_path is not None:
            return DataUrl(file_path=file_path, data_path=data_path, scheme="silx")
        else:
            return None

    @VolumeBase.data.setter
    def data(self, data):
        if not isinstance(data, (numpy.ndarray, type(None), h5py.VirtualLayout)):
            raise TypeError(
                f"data is expected to be None or a numpy array not {type(data)}"
            )
        if isinstance(data, numpy.ndarray) and data.ndim != 3:
            raise ValueError(f"data is expected to be 3D and not {data.ndim}D.")
        self._data = data

    @property
    def file_path(self):
        return self._file_path

    @file_path.setter
    def file_path(self, file_path: str | None):
        if not (file_path is None or isinstance(file_path, str)):
            raise TypeError
        self._file_path = file_path
        self.url = self._get_url_from_file_path_data_path(
            self.file_path, self.data_path
        )

    @property
    def data_path(self):
        return self._data_path

    @data_path.setter
    def data_path(self, data_path: str | None):
        if not (data_path is None or isinstance(data_path, str)):
            raise TypeError
        self._data_path = data_path
        self.url = self._get_url_from_file_path_data_path(
            self.file_path, self.data_path
        )

    @docstring(VolumeBase)
    def deduce_data_and_metadata_urls(self, url: DataUrl | None) -> tuple:
        if url is None:
            return None, None
        else:
            if url.data_slice() is not None:
                raise ValueError(f"data_slice is not handled by the {HDF5Volume}")
            file_path = url.file_path()
            data_path = url.data_path()
            if data_path is None:
                raise ValueError(
                    "data_path not provided from the DataUrl. Please provide one."
                )
            scheme = url.scheme() or "silx"
            return (
                # data url
                DataUrl(
                    file_path=file_path,
                    data_path="/".join([data_path, self.DATA_DATASET_NAME]),
                    scheme=scheme,
                ),
                # medata url
                DataUrl(
                    file_path=file_path,
                    data_path="/".join([data_path, self.METADATA_GROUP_NAME]),
                    scheme=scheme,
                ),
            )

    @docstring(VolumeBase)
    def save_data(self, url: DataUrl | None = None, mode="a", **kwargs) -> None:
        """
        :raises KeyError: if data path already exists and overwrite set to False
        :raises ValueError: if data is None
        """
        # to be discussed. Not sure we should raise an error in this case. Could be usefull but this could also be double edged knife
        if self.data is None:
            raise ValueError("No data to be saved")
        _deprecate_url_in_signature(url=url)
        url = url or self.data_url
        if url is None:
            raise ValueError(
                "Cannot get data_url. An url should be provided. Don't know where to save this."
            )
        else:
            _logger.info(f"save data to {url.path()}")

        if url.file_path() is not None and os.path.dirname(url.file_path()) != "":
            os.makedirs(os.path.dirname(url.file_path()), exist_ok=True)
        with h5py.File(url.file_path(), mode=mode) as h5s:
            if url.data_path() in h5s:
                if self.overwrite:
                    _logger.debug(
                        f"overwrite requested. Will remove {url.data_path()} entry"
                    )
                    del h5s[url.data_path()]
                else:
                    raise OSError(
                        f"Unable to save data to {url.data_path()}. This path already exists in {url.file_path()}. If you want you can ask to overwrite it."
                    )
            if isinstance(self.data, h5py.VirtualLayout):
                h5s.create_virtual_dataset(name=url.data_path(), layout=self.data)
            else:
                h5s.create_dataset(url.data_path(), data=self.data, **kwargs)

    @docstring(VolumeBase)
    def data_file_saver_generator(
        self,
        n_frames,
        data_url: DataUrl | None = None,
        overwrite: bool = False,
        mode: str = "a",
        **kwargs,
    ):
        """
        warning: the file will be open until the generator exists
        """

        class _FrameDumper:
            """
            will not work for VirtualLayout
            """

            Dataset = None
            # shared dataset

            def __init__(
                self,
                root_group,
                data_path,
                create_dataset,
                n_frames,
                i_frame,
                overwrite,
                mode,
            ) -> None:
                self.data_path = data_path
                self.root_group = root_group
                self.create_dataset = create_dataset
                self.n_frames = n_frames
                self.mode = mode
                self.overwrite = overwrite
                self.i_frame = i_frame
                self.__kwargs = kwargs  # keep chunk arguments for example

            def __setitem__(self, key, value):
                frame = value
                if _FrameDumper.Dataset is None:
                    if self.data_path in self.root_group:
                        if self.overwrite:
                            _logger.debug(
                                f"overwrite requested. Will remove {data_url.data_path()} entry"
                            )
                            del h5s[data_url.data_path()]
                        else:
                            raise OSError(
                                f"Unable to save data to {data_url.data_path()}. This path already exists in {data_url.file_path()}. If you want you can ask to overwrite it."
                            )

                    _FrameDumper.Dataset = h5s.create_dataset(  # pylint: disable=E1137
                        name=data_url.data_path(),
                        shape=(n_frames, frame.shape[0], frame.shape[1]),
                        dtype=frame.dtype,
                        **self.__kwargs,
                    )
                if key != slice(None, None, None):
                    raise ValueError("item setting only handle ':' for now")
                _FrameDumper.Dataset[i_frame] = frame  # pylint: disable=E1137

        data_url = data_url or self.data_url

        if (
            data_url.file_path() is not None
            and os.path.dirname(data_url.file_path()) != ""
        ):
            os.makedirs(os.path.dirname(data_url.file_path()), exist_ok=True)
        with h5py.File(data_url.file_path(), mode=mode) as h5s:
            for i_frame in range(n_frames):
                yield _FrameDumper(
                    create_dataset=i_frame == 0,
                    data_path=data_url.data_path(),
                    root_group=h5s,
                    n_frames=n_frames,
                    i_frame=i_frame,
                    overwrite=overwrite,
                    mode=mode,
                )

    @docstring(VolumeBase)
    def save_metadata(self, url: DataUrl | None = None) -> None:
        """
        :raises KeyError: if data path already exists and overwrite set to False
        :raises ValueError: if data is None
        """
        if self.metadata is None:
            raise ValueError("No metadata to be saved")
        _deprecate_url_in_signature(url=url)
        url = url or self.metadata_url
        if url is None:
            raise ValueError(
                "Cannot get metadata_url. An url should be provided. Don't know where to save this."
            )
        else:
            _logger.info(f"save metadata to {url.path()}")

        if url.file_path() is not None and os.path.dirname(url.file_path()) != "":
            os.makedirs(os.path.dirname(url.file_path()), exist_ok=True)

        dicttonx(
            self.metadata,
            h5file=url.file_path(),
            h5path=url.data_path(),
            update_mode="replace",
            mode="a",
        )

    @docstring(VolumeBase)
    def load_data(
        self, url: DataUrl | None = None, store: bool = True
    ) -> numpy.ndarray:
        _deprecate_url_in_signature(url=url)
        url = url or self.data_url
        if url is None:
            raise ValueError(
                "Cannot get data_url. An url should be provided. Don't know where to save this."
            )

        with open_hdf5(filename=url.file_path()) as h5s:
            if url.data_path() in h5s:
                data = h5s[url.data_path()][()]
            else:
                raise KeyError(f"Data path {url.data_path()} not found.")

        if store:
            self.data = data

        return data

    def get_slice(
        self,
        index: int | str = None,
        axis=None,
        xy=None,
        xz=None,
        yz=None,
        url: DataUrl | None = None,
    ):
        if xy is yz is xz is None and (index is None or axis is None):
            raise ValueError("index and axis should be provided")
        if xy is not None:
            deprecated_warning(
                type_="parameter",
                name="xy",
                replacement="axis and index",
            )
            if axis is None and index is None:
                axis = 0
                index = xy
            else:
                raise ValueError("several axis (previously xy, xz, yz requested")
        elif xz is not None:
            deprecated_warning(
                type_="parameter",
                name="xz",
                replacement="axis and index",
            )
            if axis is None and index is None:
                axis = 1
                index = xz
            else:
                raise ValueError("several axis (previously xy, xz, yz requested")
        elif yz is not None:
            deprecated_warning(
                type_="parameter",
                name="yz",
                replacement="axis and index",
            )
            if axis is None and index is None:
                axis = 2
                index = yz
            else:
                raise ValueError("several axis (previously xy, xz, yz requested")

        if isinstance(index, str):
            if index == "first":
                index = 0
            elif index == "middle":
                index = self.get_volume_shape()[axis] // 2
            elif index == "last":
                index = -1
            else:
                raise ValueError(f"index '{index}' is not handled")

        if self.data is not None:
            return self.select(volume=self.data, axis=axis, index=index)
        else:
            url = url or self.data_url
            if url is None:
                raise ValueError(
                    "Cannot get data_url. An url should be provided. Don't know where to save this."
                )
            with open_hdf5(filename=url.file_path()) as h5s:
                if url.data_path() in h5s:
                    return self.select(
                        volume=h5s[url.data_path()], axis=axis, index=index
                    )
                else:
                    raise KeyError(f"Data path {url.data_path()} not found.")

    def _get_slices_from_disk(
        self, slices: tuple[SliceTuple], url: DataUrl | None = None
    ) -> dict[SliceTuple, numpy.ndarray]:
        """
        read from files a couple of slices along any axis.
        """
        _deprecate_url_in_signature(url=url)
        url = url or self.data_url
        with open_hdf5(filename=url.file_path()) as h5s:
            if url.data_path() in h5s:
                return self.select_slices(volume=h5s[url.data_path()], slices=slices)

    @docstring(VolumeBase)
    def load_metadata(self, url: DataUrl | None = None, store: bool = True) -> dict:
        _deprecate_url_in_signature(url=url)
        url = url or self.metadata_url
        if url is None:
            raise ValueError(
                "Cannot get metadata_url. An url should be provided. Don't know where to save this."
            )
        try:
            metadata = nxtodict(
                h5file=url.file_path(), path=url.data_path(), asarray=False
            )
        except KeyError:
            _logger.warning(f"no metadata found in {url.data_path()}")
            metadata = {}
        if store:
            self.metadata = metadata
        return metadata

    def browse_metadata_files(self, url=None):
        """
        return a generator go through all the existing files associated to the data volume
        """
        _deprecate_url_in_signature(url=url)
        url = url or self.metadata_url
        if url is None:
            return
        elif url.file_path() is not None and os.path.exists(url.file_path()):
            yield url.file_path()

    def browse_data_files(self, url=None):
        """
        return a generator go through all the existing files associated to the data volume
        """
        _deprecate_url_in_signature(url=url)
        url = url or self.data_url
        if url is None:
            return
        elif url.file_path() is not None and os.path.exists(url.file_path()):
            yield url.file_path()

    def browse_data_urls(self, url=None):
        _deprecate_url_in_signature(url=url)
        url = url or self.data_url
        if url is not None and os.path.exists(url.file_path()):
            yield url

    @staticmethod
    @docstring(VolumeBase)
    def from_identifier(identifier):
        """Return the Dataset from a identifier"""
        if not isinstance(identifier, HDF5VolumeIdentifier):
            raise TypeError(
                f"identifier should be an instance of {HDF5VolumeIdentifier}"
            )
        return HDF5Volume(
            file_path=identifier.file_path,
            data_path=identifier.data_path,
        )

    @docstring(VolumeBase)
    def get_identifier(self) -> HDF5VolumeIdentifier:
        if self.url is None:
            raise ValueError("no file_path provided. Cannot provide an identifier")
        return HDF5VolumeIdentifier(
            object=self, hdf5_file=self.url.file_path(), entry=self.url.data_path()
        )

    @staticmethod
    def example_defined_from_str_identifier() -> str:
        return (
            HDF5Volume(file_path="/path/to/file_path", data_path="entry0000")
            .get_identifier()
            .to_str()
        )

    @docstring(VolumeBase)
    def browse_slices(self, url=None):
        _deprecate_url_in_signature(url=url)
        if url is None and self.data is not None:
            for data_slice in self.data:
                yield data_slice
        else:
            url = url or self.data_url
            if url is None:
                raise ValueError(
                    "No data and data_url know and no url provided. Uanble to browse slices"
                )
            with open_hdf5(filename=url.file_path()) as h5s:
                if url.data_path() in h5s:
                    for data_slice in h5s[url.data_path()]:
                        yield data_slice
                else:
                    raise KeyError(f"Data path {url.data_path()} not found.")

    @docstring(VolumeBase)
    def load_chunk(self, chunk, url=None):
        _deprecate_url_in_signature(url=url)
        url = url or self.data_url
        if url is None:
            raise ValueError("Cannot get data_url. An url should be provided.")
        with open_hdf5(filename=url.file_path()) as h5s:
            if url.data_path() in h5s:
                return h5s[url.data_path()][chunk]
            else:
                raise KeyError(f"Data path {url.data_path()} not found.")

    def get_volume_shape(self, url=None):
        _deprecate_url_in_signature(url=url)
        if self.data is not None:
            return self.data.shape
        url = url or self.data_url
        if url is None:
            raise ValueError("Cannot get data_url. An url should be provided.")
        else:
            with open_hdf5(filename=url.file_path()) as h5s:
                if url.data_path() in h5s:
                    return h5s[url.data_path()].shape
                else:
                    return None


def get_default_data_path_for_volume(scan: TomoScanBase) -> str:
    if not isinstance(scan, TomoScanBase):
        raise TypeError(
            f"scan is expected to be an instance of {TomoScanBase} not {type(scan)}"
        )
    entry = getattr(scan, "entry", "entry")
    return "/".join([entry, "reconstruction"])



# coding: utf-8
from __future__ import annotations

import os

import fabio
import numpy

from tomoscan.esrf.mock import MockEDF as _MockEDF
from tomoscan.factory import Factory
from tomoscan.scanbase import ReducedFramesInfos


class MockEDFWithRawRef(_MockEDF):
    def __init__(
        self,
        scan_path,
        n_radio,
        flats: dict,
        darks: dict,
        n_extra_radio=0,
        scan_range=360,
        n_recons=0,
        n_pag_recons=0,
        recons_vol=False,
        dim=200,
        scene="noise",
        machine_current: numpy.array | None = None,
        count_time: numpy.array | None = None,
    ):
        self.flat_n = len(flats[list(flats.keys())[0]]) if len(flats) > 0 else 0
        self.dark_n = len(flats[list(darks.keys())[0]]) if len(darks) > 0 else 0
        self.flats = flats
        self.darks = darks
        self.dim = dim
        self._machine_current = machine_current
        self._count_time = count_time
        super().__init__(
            scan_path,
            n_radio,
            n_ini_radio=n_radio,
            n_extra_radio=n_extra_radio,
            scan_range=scan_range,
            n_recons=n_recons,
            n_pag_recons=n_pag_recons,
            recons_vol=recons_vol,
            dim=dim,
            scene=scene,
            dark_n=self.dark_n,
            flat_n=self.flat_n,
        )

        # add some raw dark and raw flat
        self.add_darks()
        self.add_flats()

        # add some darkHST and rehHST containing other data...
        self.addFalseReducedFlats()
        self.addFalseReducedDark()

    def add_dark(self, index):
        pass

    def add_flat(self, index):
        pass

    def add_flats(self):
        for i_serie, flats in self.flats.items():
            for i_flat, flat in enumerate(flats):
                file_name = f"ref{str(i_flat).zfill(4)}_{str(i_serie).zfill(4)}.edf"
                file_path = os.path.join(self.scan_path, file_name)
                if not os.path.exists(file_path):
                    header = {
                        "motor_pos": f"{i_serie} 0.0 1.0 2.0",
                        "motor_mne": "srot sx sy sz",
                    }
                    edf_writer = fabio.edfimage.EdfImage(
                        data=flat,
                        header=header,
                    )
                    edf_writer.write(file_path)

    def add_darks(self):
        for i_serie, darks in self.darks.items():
            assert darks.ndim == 3
            assert darks.shape[0] == 1
            file_name = f"darkend{i_serie:04}.edf"
            file_path = os.path.join(self.scan_path, file_name)
            if not os.path.exists(file_path):
                edf_writer = fabio.edfimage.EdfImage(
                    data=darks.reshape(darks.shape[1], darks.shape[2]),
                    header={
                        "motor_pos": f"{i_serie} 0.0 1.0 2.0",
                        "motor_mne": "srot sx sy sz",
                    },
                )
                edf_writer.write(file_path)

    def addFalseReducedDark(self):
        data = numpy.zeros((self.dim, self.dim)) - 1
        file_name = "dark.edf"
        file_path = os.path.join(self.scan_path, file_name)
        if not os.path.exists(file_path):
            edf_writer = fabio.edfimage.EdfImage(
                data=data,
                header={
                    "motor_pos": f"{-1} 0.0 1.0 2.0",
                    "motor_mne": "srot sx sy sz",
                },
            )
            edf_writer.write(file_path)

    def addFalseReducedFlats(self):
        for i_serie, flats in self.flats.items():
            data = numpy.zeros((self.dim, self.dim)) - 1
            file_name = f"refHST{i_serie:04}.edf"
            file_path = os.path.join(self.scan_path, file_name)
            if not os.path.exists(file_path):
                edf_writer = fabio.edfimage.EdfImage(
                    data=data,
                    header={
                        "motor_pos": f"{-1} 0.0 1.0 2.0",
                        "motor_mne": "srot sx sy sz",
                    },
                )
            edf_writer.write(file_path)

    def write_metadata(self, n_radio, scan_range, flat_n, dark_n):
        super().write_metadata(
            n_radio=n_radio, scan_range=scan_range, flat_n=flat_n, dark_n=dark_n
        )
        info_file = self.get_info_file()
        with open(info_file, "a") as info_file:
            if self._machine_current is not None:
                info_file.write("SrCurrent=    " + str(self._machine_current) + "\n")
            if self._count_time is not None:
                info_file.write("Count_time=    " + str(self._count_time) + "\n")


def test_reduce_edf(tmp_path):
    dim = 20
    folder_1 = tmp_path / "test1"
    folder_1.mkdir()
    n_proj = 12
    n_darks = 1

    count_time = 1.2  # for edf now we only consider a scalar (store in the .info file)
    machine_current = (
        10.2  # for edf now we only consider a scalar (store in the .info file)
    )

    raw_darks = numpy.ones((n_darks, dim, dim), dtype="f")

    raw_flats_s1 = numpy.asarray(
        [
            numpy.ones((dim, dim), dtype=numpy.float32),
            numpy.ones((dim, dim), dtype=numpy.float32) + 1.0,
            numpy.ones((dim, dim), dtype=numpy.float32) + 2.0,
            numpy.ones((dim, dim), dtype=numpy.float32) + 3.0,
            numpy.ones((dim, dim), dtype=numpy.float32) + 4.0,
        ]
    )

    raw_flats_s2 = numpy.asarray(
        [
            numpy.ones((dim, dim), dtype=numpy.float32) + 10.0,
            numpy.ones((dim, dim), dtype=numpy.float32) + 11.0,
            numpy.ones((dim, dim), dtype=numpy.float32) + 12.0,
            numpy.ones((dim, dim), dtype=numpy.float32) + 13.0,
            numpy.ones((dim, dim), dtype=numpy.float32) + 14.0,
        ]
    )

    MockEDFWithRawRef(
        scan_path=str(folder_1),
        n_radio=n_proj,
        flats={0: raw_flats_s1, 12: raw_flats_s2},
        darks={0: raw_darks},
        machine_current=machine_current,
        count_time=count_time,
    )
    scan = Factory.create_scan_object(str(folder_1))

    # test reduced frames

    numpy.testing.assert_array_equal(
        scan.compute_reduced_darks(reduced_method="median")[0],
        scan.compute_reduced_darks(reduced_method="mean")[0],
    )

    numpy.testing.assert_array_equal(
        scan.compute_reduced_flats(reduced_method="median")[0],
        numpy.median(raw_flats_s1, axis=0).astype(numpy.int32),
    )
    numpy.testing.assert_array_equal(
        scan.compute_reduced_flats(reduced_method="mean", output_dtype=numpy.float32)[
            0
        ],
        numpy.mean(raw_flats_s1, axis=0),
    )

    numpy.testing.assert_array_equal(
        scan.compute_reduced_flats(reduced_method="median")[12],
        numpy.median(raw_flats_s2, axis=0).astype(numpy.int32),
    )
    numpy.testing.assert_array_equal(
        scan.compute_reduced_flats(reduced_method="mean")[12],
        numpy.mean(raw_flats_s2, axis=0).astype(numpy.int32),
    )

    numpy.testing.assert_array_equal(
        scan.compute_reduced_flats(reduced_method="first")[0],
        numpy.ones((dim, dim), dtype=numpy.int32),
    )

    numpy.testing.assert_array_equal(
        scan.compute_reduced_flats(reduced_method="last")[0],
        numpy.ones((dim, dim), dtype=numpy.int32) + 4,
    )

    numpy.testing.assert_array_equal(
        scan.compute_reduced_flats(reduced_method="first", output_dtype=None)[12],
        numpy.ones((dim, dim), dtype=numpy.float32) + 10.0,
    )

    # test reduced metadata
    _, darks_metadata = scan.compute_reduced_darks(
        reduced_method="median", return_info=True
    )
    assert isinstance(darks_metadata, ReducedFramesInfos)
    assert len(darks_metadata.count_time) == len(darks_metadata.machine_current) == 1
    # with the current method (only one Srcurrent and Count_time value per acquistion) flats and darks will have the same value
    # and also the same no matter the method to compute it
    assert darks_metadata.count_time[0] == count_time
    assert darks_metadata.machine_current[0] == machine_current
    _, flats_metadata = scan.compute_reduced_flats(
        reduced_method="mean", return_info=True
    )
    assert len(flats_metadata.count_time) == len(flats_metadata.machine_current) == 2
    assert (
        darks_metadata.count_time[0]
        == flats_metadata.count_time[0]
        == flats_metadata.count_time[1]
    )
    assert (
        darks_metadata.machine_current[0]
        == flats_metadata.machine_current[0]
        == flats_metadata.machine_current[1]
    )


def test_reduce_edf_fails(tmp_path):
    folder_1 = tmp_path / "test2"

    MockEDFWithRawRef(scan_path=str(folder_1), n_radio=12, flats={}, darks={})
    scan = Factory.create_scan_object(str(folder_1))
    assert scan.compute_reduced_flats(reduced_method="first", return_info=False) == {}
    assert scan.compute_reduced_darks(reduced_method="last", return_info=False) == {}



# coding: utf-8

import numpy


class UrlSettings:
    FILE_PATH_KEY = "file_path"
    DATA_PATH_KEY = "path"
    FILE_PREFIX = "file_prefix"


def split_query(query: str) -> dict:
    result = dict()
    for s in query.split("&"):
        if not s:
            continue
        name, _, value = s.partition("=")
        prev_value = result.get(name)
        if prev_value:
            value = join_string(prev_value, value, "/")
        result[name] = value
    return result


def join_query(
    query_items: list[tuple[str, str]] | tuple[tuple[str, str], ...] | numpy.ndarray,
) -> str:
    return "&".join(f"{k}={v}" for k, v in query_items)


def join_string(a: str, b: str, sep: str):
    aslash = a.endswith(sep)
    bslash = b.startswith(sep)
    if aslash and bslash:
        return a[:-1] + b
    elif aslash or bslash:
        return a + b
    else:
        return a + sep + b


def join_path(path_items: tuple) -> str:
    if not isinstance(path_items, tuple):
        raise TypeError
    return ":".join(path_items)


def split_path(path: str):
    return path.split(":")



# coding: utf-8
"""
Test of the tiffmodule.
Note: some test are also perform on the test_single_frame_volume
"""


import os

import numpy
import pytest
from silx.io.url import DataUrl
from tifffile import TiffFile

from tomoscan.esrf.identifier.tiffidentifier import MultiTiffVolumeIdentifier
from tomoscan.esrf.volume.mock import create_volume
from tomoscan.esrf.volume.tiffvolume import MultiTIFFVolume, TIFFVolume
from tomoscan.identifier import VolumeIdentifier

_data = create_volume(frame_dims=(20, 20), z_size=5)
_data = _data.astype(numpy.uint8)

_metadata = {
    "nabu_config": {
        "dataset": {
            "location": "toto.hdf5",
            "entry": "entry0000",
        },
        "phase": {
            "method": "None",
            "padding_type": "edge",
        },
    },
    "processing_option": {
        "build_sino": {
            "axis_correction": "None",
            "enable_halftomo": True,
        },
        "flatfield": {
            "binning": "11",
            "do_flat_distortion": False,
        },
    },
}


def test_multi_tiff_volume(tmp_path):
    test_dir = str(tmp_path / "test_volume_url")
    os.makedirs(test_dir)
    volume_file = os.path.join(test_dir, "volume.tiff")

    volume = MultiTIFFVolume(
        file_path=volume_file,
        data=_data,
        metadata=_metadata,
    )

    assert volume.url.file_path() == volume_file
    assert volume.url.data_path() is None
    assert volume.data_url.scheme() == "tifffile"
    assert volume.data_url.file_path() == volume_file
    assert volume.data_url.data_path() is None
    assert volume.metadata_url.scheme() == "ini"
    expected_metadata_file = os.path.join(test_dir, "volume_infos.txt")
    assert volume.metadata_url.file_path() == expected_metadata_file
    assert volume.metadata_url.data_path() is None

    volume_identifier = volume.get_identifier()
    assert isinstance(volume_identifier.short_description(), str)
    assert isinstance(volume_identifier, VolumeIdentifier)

    assert not os.path.exists(volume_file)
    assert tuple(volume.browse_data_files()) == tuple()
    assert tuple(volume.browse_data_urls()) == tuple()
    assert tuple(volume.browse_metadata_files()) == tuple()

    volume.save()
    assert os.path.exists(volume_file)
    assert tuple(volume.browse_metadata_files()) == (expected_metadata_file,)
    assert tuple(volume.browse_data_files()) == (volume_file,)
    assert len(list(volume.browse_data_urls())) == len((volume.data_url,))

    volume_loaded = MultiTIFFVolume(
        file_path=volume_file,
    )
    assert volume_loaded.metadata_url.file_path() == expected_metadata_file
    assert volume_loaded.metadata_url.data_path() is None

    assert volume_loaded._data is None
    assert volume_loaded._metadata is None
    # check loading function won't save loaded data in cache
    data_no_cache = volume_loaded.load_data(volume_loaded.data_url, store=False)
    metadata_no_cache = volume_loaded.load_metadata(
        volume_loaded.metadata_url, store=False
    )
    assert data_no_cache is not None
    assert metadata_no_cache is not None
    assert volume_loaded._data is None
    assert volume_loaded._metadata is None
    numpy.testing.assert_array_almost_equal(_data, data_no_cache)
    assert _metadata == metadata_no_cache
    assert volume.get_min_max_values() == (_data.min(), _data.max())

    # check data and metadata properties will load and set data in cache
    volume_loaded.load()
    data_cache = volume_loaded.data
    assert data_cache is not None
    metadata_cache = volume_loaded.metadata
    assert metadata_cache is not None
    numpy.testing.assert_array_almost_equal(_data, data_cache)
    assert _metadata == metadata_cache
    volume_loaded.data = None
    volume_loaded.metadata = None
    with pytest.raises(ValueError):
        volume_loaded.save_data()

    with pytest.raises(ValueError):
        volume_loaded.save_metadata()

    with pytest.raises(TypeError):
        volume_loaded.data = "toto"

    volume_loaded.data = create_volume(frame_dims=(20, 20), z_size=2)
    with pytest.raises(ValueError):
        volume_loaded.data = create_volume(frame_dims=(20, 20), z_size=2)[0]

    # get the volume from the identifier
    volume_from_id = MultiTIFFVolume.from_identifier(identifier=volume_identifier)
    assert volume_from_id.data is None
    assert volume_from_id.metadata is None

    # get the volume from the identifier str
    volume_from_str = MultiTIFFVolume.from_identifier(
        identifier=MultiTiffVolumeIdentifier.from_str(volume_identifier.to_str())
    )
    assert volume_from_str.data is None
    assert volume_from_str.metadata is None
    assert volume_from_str.get_identifier() == volume_from_id.get_identifier()
    assert volume_from_str.get_identifier() != "toto"

    # test browsing frames
    for l_frame, o_frame in zip(volume_from_str.browse_slices(), _data):
        numpy.testing.assert_allclose(l_frame, o_frame)

    # test hash
    hash(volume_from_str)

    # test append param
    volume_file_append = os.path.join(test_dir, "volume_append.tiff")
    volume_append = MultiTIFFVolume(
        file_path=volume_file_append, data=_data, metadata=_metadata, append=True
    )
    volume_append.data = create_volume(frame_dims=(20, 20), z_size=2)
    volume_append.save()
    volume_append2 = MultiTIFFVolume(
        file_path=volume_file_append, data=_data, metadata=_metadata, append=True
    )
    volume_append2.data = create_volume(frame_dims=(20, 20), z_size=3) + 10
    volume_append2.save()
    tiff_append = TiffFile(volume_file_append)
    assert len(tiff_append.series) == 5

    # test load_chunk
    chunks_to_test = [
        (slice(None, None), slice(0, 5), slice(None, None)),
        (slice(0, 4), slice(0, 15), slice(None, None)),
    ]
    for chunk in chunks_to_test:
        data_chunk = volume.load_chunk(chunk)
        assert data_chunk.shape == _data[chunk].shape
        assert numpy.allclose(data_chunk, _data[chunk])
    assert volume.data_extension == "tiff"
    assert volume.metadata_extension == "txt"


def test_data_file_saver_generator(tmp_path):
    """
    data_file_saver_generator (dumping frame by frame a volume to disk)
    """
    volume_file_path = str(tmp_path / "volume.h5")
    volume = MultiTIFFVolume(file_path=volume_file_path)
    for slice_, slice_saver in zip(
        _data,
        volume.data_file_saver_generator(
            n_frames=_data.shape[0], data_url=volume.data_url, overwrite=False
        ),
    ):
        slice_saver[:] = slice_
    assert volume.data is None
    numpy.testing.assert_array_equal(
        volume.load_data(),
        _data,
    )


def test_tiff_volume_nabu_single_frame_pattern(tmp_path):
    """
    Insure we can save and load using a specific pattermn (single frame volume) that can happen in nabu for example
    """
    test_dir = str(tmp_path / "test_volume_url")
    os.makedirs(test_dir)
    output_file = os.path.join(test_dir, "frame.tiff")

    class SingleFrameTIFFVolume(TIFFVolume):
        # we are not insure that the output directory name is the base name of the file_path
        DEFAULT_DATA_PATH_PATTERN = "frame.tiff"

    frame = _data[0].reshape(1, _data.shape[1], _data.shape[2])
    volume = SingleFrameTIFFVolume(
        folder=test_dir,
        data=frame,
        metadata=_metadata,
        overwrite=True,
    )
    volume.save()
    assert os.path.exists(output_file)

    volume_loaded_1 = SingleFrameTIFFVolume(
        folder=test_dir,
    )
    volume_loaded_1.load()
    numpy.testing.assert_array_equal(
        frame,
        volume_loaded_1.data,
    )

    volume_loaded_2 = TIFFVolume(
        data_url=DataUrl(
            file_path=output_file,
            data_path="frame.tiff",
            scheme="tifffile",
        ),
    )
    volume_loaded_2.load_data(store=True)
    numpy.testing.assert_array_equal(
        frame,
        volume_loaded_2.data,
    )

    # test get_slice function
    with pytest.raises(IndexError):
        numpy.testing.assert_array_equal(
            volume_loaded_2.get_slice(index=2, axis=0),
            _data[2],
        )
    numpy.testing.assert_array_equal(
        volume_loaded_2.get_slice(index=0, axis=0),
        _data[0],
    )
    numpy.testing.assert_array_equal(
        volume_loaded_2.get_slice(index=2, axis=1),
        frame[:, 2, :],
    )
    numpy.testing.assert_array_equal(
        volume_loaded_2.get_slice(index=2, axis=2),
        frame[:, :, 2],
    )
    volume_loaded_2.data = None

    numpy.testing.assert_array_equal(
        volume_loaded_2.get_slice(index=2, axis=1),
        frame[:, 2, :],
    )
    numpy.testing.assert_array_equal(
        volume_loaded_2.get_slice(index=2, axis=2),
        frame[:, :, 2],
    )


def test_tiff_volume_nabu_multi_frame_pattern(tmp_path):
    """
    Insure we can save and load using a specific pattermn (multi frame volume) that can happen in nabu for example
    """
    test_dir = str(tmp_path / "test_volume_url")
    os.makedirs(test_dir)

    class MultiFrameTIFFVolume(TIFFVolume):
        # we are not insure that the output directory name is the base name of the file_path
        DEFAULT_DATA_PATH_PATTERN = "frame_{index_zfill6}.tiff"

    volume = MultiFrameTIFFVolume(
        folder=test_dir,
        data=_data,
        metadata=_metadata,
        overwrite=True,
    )
    volume.save()
    output_file_first_frame = os.path.join(test_dir, "frame_000000.tiff")
    assert os.path.exists(output_file_first_frame)

    volume_loaded_1 = MultiFrameTIFFVolume(
        folder=test_dir,
    )
    volume_loaded_1.load()
    numpy.testing.assert_array_equal(
        _data,
        volume_loaded_1.data,
    )

    volume_loaded_2 = TIFFVolume(
        data_url=DataUrl(
            file_path=test_dir,
            data_path="frame_{index_zfill6}.tiff",
            scheme="tifffile",
        ),
    )
    volume_loaded_2.load_data(store=True)
    numpy.testing.assert_array_equal(
        _data,
        volume_loaded_2.data,
    )

    # test get_slice function
    numpy.testing.assert_array_equal(
        volume_loaded_2.get_slice(index=2, axis=0),
        _data[2],
    )
    numpy.testing.assert_array_equal(
        volume_loaded_2.get_slice(index=2, axis=1),
        _data[:, 2, :],
    )
    numpy.testing.assert_array_equal(
        volume_loaded_2.get_slice(index=2, axis=2),
        _data[:, :, 2],
    )
    volume_loaded_2.data = None

    numpy.testing.assert_array_equal(
        volume_loaded_2.get_slice(index=2, axis=0),
        _data[2],
    )
    numpy.testing.assert_array_equal(
        volume_loaded_2.get_slice(index=2, axis=1),
        _data[:, 2, :],
    )
    numpy.testing.assert_array_equal(
        volume_loaded_2.get_slice(index=2, axis=2),
        _data[:, :, 2],
    )


def test_example():
    """test static function 'example'"""
    assert isinstance(MultiTIFFVolume.example_defined_from_str_identifier(), str)



from __future__ import annotations


import numpy
import warnings


class ReducedFramesInfos:
    """contains reduced frames metadata as count_time and machine_current"""

    MACHINE_ELECT_CURRENT_KEY = "machine_current"

    COUNT_TIME_KEY = "count_time"

    LR_FLIP = "lr_flip"
    """Save the information if the original raw frame was left-right flipped. Warning: tomoscan will only propagate this information but won't flip reduced frames saved"""

    UD_FLIP = "ud_flip"
    """Save the information if the original raw frame was up-down flipped. Warning: tomoscan will only propagate this information but won't flip reduced frames saved"""

    def __init__(self) -> None:
        self._count_time = []
        self._machine_current = []
        self._lr_flip = None
        self._ud_flip = None

    def __eq__(self, __o: object) -> bool:
        if isinstance(__o, dict):
            return ReducedFramesInfos().load_from_dict(__o) == self
        if not isinstance(__o, ReducedFramesInfos):
            return False
        return numpy.array_equal(
            numpy.array(self.count_time), numpy.array(__o.count_time)
        ) and numpy.array_equal(
            numpy.array(self.machine_current),
            numpy.array(__o.machine_current),
        )

    def clear(self):
        self._count_time.clear()
        self._machine_current.clear()
        self._lr_flip = None
        self._ud_flip = None

    @property
    def count_time(self) -> list:
        """
        frame exposure time in second
        """
        return self._count_time

    @count_time.setter
    def count_time(self, count_time: list | tuple | numpy.ndarray | None):
        if count_time is None:
            self._count_time.clear()
        else:
            self._count_time = list(count_time)

    @property
    def machine_electric_current(self) -> list:
        warnings.warn(
            "machine_electric_current is deprecated and will be removed in a future version. Use machine_current instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        return self.machine_current

    @machine_electric_current.setter
    def machine_electric_current(
        self, machine_current: list | tuple | numpy.ndarray | None
    ):
        warnings.warn(
            "machine_electric_current is deprecated and will be removed in a future version. Use machine_current instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        self.machine_current = machine_current

    @property
    def machine_current(self) -> list:
        """
        machine electric current in Ampere
        """
        return self._machine_current

    @machine_current.setter
    def machine_current(self, machine_current: list | tuple | numpy.ndarray | None):
        if machine_current is None:
            self._machine_current.clear()
        else:
            self._machine_current = list(machine_current)

    @property
    def lr_flip(self) -> bool | None:
        return self._lr_flip

    @lr_flip.setter
    def lr_flip(self, flip: bool | None):
        if not isinstance(flip, (bool, type(None))):
            raise TypeError(
                f"flip is expected to be None or a bool. got {type(flip)}, {flip}, {flip.dtype}"
            )
        self._lr_flip = flip

    @property
    def ud_flip(self) -> bool | None:
        return self._ud_flip

    @ud_flip.setter
    def ud_flip(self, flip: bool | None):
        if not isinstance(flip, (bool, type(None))):
            raise TypeError(
                f"flip is expected to be None or a bool. got {type(flip)}, {flip}, {flip.dtype}"
            )
        self._ud_flip = flip

    def to_dict(self) -> dict:
        res = {}
        if len(self.machine_current) > 0:
            res[self.MACHINE_ELECT_CURRENT_KEY] = self.machine_current
        if len(self.count_time) > 0:
            res[self.COUNT_TIME_KEY] = self.count_time
        if self.lr_flip is not None:
            res[self.LR_FLIP] = self.lr_flip
        if self.ud_flip is not None:
            res[self.UD_FLIP] = self.ud_flip
        return res

    def load_from_dict(self, my_dict: dict):
        self.machine_current = my_dict.get(self.MACHINE_ELECT_CURRENT_KEY, None)
        self.count_time = my_dict.get(self.COUNT_TIME_KEY, None)

        def cast_numpy_bool(value):
            if isinstance(value, numpy.ndarray):
                return bool(value)
            return value

        self.lr_flip = cast_numpy_bool(my_dict.get(self.LR_FLIP, None))
        self.ud_flip = cast_numpy_bool(my_dict.get(self.UD_FLIP, None))
        return self

    @staticmethod
    def pop_info_keys(my_dict: dict):
        if not isinstance(my_dict, dict):
            raise TypeError
        my_dict.pop(ReducedFramesInfos.MACHINE_ELECT_CURRENT_KEY, None)
        my_dict.pop(ReducedFramesInfos.COUNT_TIME_KEY, None)
        my_dict.pop(ReducedFramesInfos.LR_FLIP, None)
        my_dict.pop(ReducedFramesInfos.UD_FLIP, None)
        return my_dict

    @staticmethod
    def split_data_and_metadata(my_dict):
        metadata = ReducedFramesInfos().load_from_dict(my_dict)
        data = ReducedFramesInfos.pop_info_keys(my_dict)

        def cast_keys_to_int(key):
            try:
                return int(key)
            except ValueError:
                return key

        data = {cast_keys_to_int(key): value for key, value in data.items()}
        return data, metadata

    def __str__(self):
        return "\n".join(
            [
                f"machine_current {self.machine_current}",
                f"count_time {self.count_time}",
                f"lr_flip {self.lr_flip}",
                f"ud_flip {self.ud_flip}",
            ]
        )



# coding: utf-8

import os
from copy import deepcopy

import numpy
import pytest
import h5py
import resource
from silx.io.dictdump import dicttoh5
from silx.io.url import DataUrl

from nxtomo.application.nxtomo import ImageKey

from tomoscan.esrf.scan.edfscan import EDFTomoScan
from tomoscan.esrf.scan.nxtomoscan import NXtomoScan
from tomoscan.esrf.scan.mock import MockEDF, MockNXtomo
from tomoscan.esrf.scan.utils import (
    check_possible_issue_with_rlimit,
    copy_h5_dict_darks_to,
    copy_h5_dict_flats_to,
    from_absolute_reduced_frames_to_relative,
    from_relative_reduced_frames_to_absolute,
    from_bliss_original_file_to_raw,
    get_files_from_pattern,
    get_series_slice,
    get_n_series,
    get_parameters_frm_par_or_info,
)
from tomoscan.tests.utils import NXtomoMockContext
from tomoscan.scanbase import ReducedFramesInfos


def test_get_files_from_index_pattern(tmp_path):
    """
    Insure we can properly find files with the `index` pattern as
    myfile{index}.edf
    """
    test_dir = tmp_path / "test_dir"
    os.makedirs(test_dir, exist_ok=True)

    # create files to be tested
    indexes = (0, 10, 100, 1000, 2025)
    for index in indexes:
        open(os.path.join(test_dir, f"file{index}.EDF"), "w").close()
    # add some noise for test
    open(os.path.join(test_dir, "file.EDF"), "w").close()
    open(os.path.join(test_dir, "tata.edf"), "w").close()
    open(os.path.join(test_dir, "file3.hdf5"), "w").close()

    found_files = get_files_from_pattern(
        file_pattern="file{index}.edf",
        pattern="index",
        research_dir=str(test_dir),
    )

    assert len(found_files) == len(indexes)
    for index in indexes:
        assert index in found_files


def test_get_files_from_index_zfill4_pattern(tmp_path):
    """
    Insure we can properly find files with the `index` pattern as
    myfile{index_zfill4}.edf
    """
    test_dir = tmp_path / "test_dir"
    os.makedirs(test_dir, exist_ok=True)

    indexes = (0, 10, 100, 1000, 2025)
    for index in indexes:
        index_zfill4 = str(index).zfill(4)
        open(os.path.join(test_dir, f"file{index_zfill4}.edf"), "w").close()

    # add some noise for test
    open(os.path.join(test_dir, "file.EDF"), "w").close()
    open(os.path.join(test_dir, "tata0000.edf"), "w").close()
    open(os.path.join(test_dir, "file0003.hdf5"), "w").close()

    found_files = get_files_from_pattern(
        file_pattern="file{index}.edf",
        pattern="index",
        research_dir=str(test_dir),
    )

    assert len(found_files) == len(indexes)
    for index in indexes:
        assert index in found_files


def test_dark_and_flat_copy(tmp_path):
    """
    test copy_darks_to and copy_flats_to functions
    """
    scan_path = os.path.join(tmp_path, "my_acqui")
    scan = MockNXtomo(
        scan_path=scan_path,
        dim=100,
        n_proj=1,
        n_ini_proj=1,
        create_ini_flat=False,
        create_ini_dark=False,
        create_final_flat=False,
    ).scan

    # create darks and flats
    darks = {
        0: numpy.ones((100, 100), dtype=numpy.float32),
    }
    flats = {
        1: numpy.ones((100, 100), dtype=numpy.float32) * 2.0,
        100: numpy.ones((100, 100), dtype=numpy.float32) * 2.0,
    }

    original_dark_flat_file = os.path.join(tmp_path, "originals.hdf5")

    dicttoh5(darks, h5file=original_dark_flat_file, h5path="darks", mode="a")
    dicttoh5(flats, h5file=original_dark_flat_file, h5path="flats", mode="a")

    darks_url = DataUrl(
        file_path=original_dark_flat_file,
        data_path="/darks",
        scheme="silx",
    )
    flats_url = DataUrl(
        file_path=original_dark_flat_file,
        data_path="/flats",
        scheme="silx",
    )

    # check copy is correctly made
    expected_dark_file = os.path.join(scan.path, "my_acqui_darks.hdf5")
    expected_flat_file = os.path.join(scan.path, "my_acqui_flats.hdf5")
    assert not os.path.exists(expected_dark_file)
    assert not os.path.exists(expected_flat_file)
    copy_h5_dict_darks_to(scan=scan, darks_url=darks_url, save=False)
    assert not os.path.exists(expected_dark_file)
    assert not os.path.exists(expected_flat_file)
    copy_h5_dict_darks_to(scan=scan, darks_url=darks_url, save=True)
    assert os.path.exists(expected_dark_file)
    assert not os.path.exists(expected_flat_file)
    copy_h5_dict_flats_to(scan=scan, flats_url=flats_url, save=False)
    assert not os.path.exists(expected_flat_file)
    copy_h5_dict_flats_to(scan=scan, flats_url=flats_url, save=True)
    assert os.path.exists(expected_flat_file)

    # check a new scan can retrieve darks and flats
    new_scan = NXtomoScan(scan_path, scan.entry)
    assert new_scan.load_reduced_darks().keys() == darks.keys()
    assert new_scan.load_reduced_flats().keys() == flats.keys()

    # check behavior if DataUrl lead to no data:
    url = DataUrl(file_path=None, data_path=None)
    with pytest.raises(Exception):
        copy_h5_dict_darks_to(scan=scan, darks_url=url, raise_error_if_url_empty=True)
    with pytest.raises(Exception):
        copy_h5_dict_flats_to(scan=scan, flats_url=url, raise_error_if_url_empty=True)

    copy_h5_dict_darks_to(scan=scan, darks_url=url, raise_error_if_url_empty=False)
    copy_h5_dict_flats_to(scan=scan, flats_url=url, raise_error_if_url_empty=False)

    # test metadata
    darks_with_metadata = deepcopy(darks)
    darks_with_metadata[ReducedFramesInfos.COUNT_TIME_KEY] = [1.0]
    darks_with_metadata[ReducedFramesInfos.MACHINE_ELECT_CURRENT_KEY] = [12.3]
    dicttoh5(
        darks_with_metadata, h5file=original_dark_flat_file, h5path="darks", mode="a"
    )

    flats_with_metadata = deepcopy(flats)
    flats_with_metadata[ReducedFramesInfos.COUNT_TIME_KEY] = [2.0, 2.1]
    flats_with_metadata[ReducedFramesInfos.MACHINE_ELECT_CURRENT_KEY] = [14.3, 14.2]
    dicttoh5(
        flats_with_metadata, h5file=original_dark_flat_file, h5path="flats", mode="a"
    )

    scan.set_reduced_darks({})
    scan.set_reduced_flats({})
    scan.reduced_darks_infos = None
    scan.reduced_flats_infos = None
    copy_h5_dict_darks_to(
        scan=scan,
        darks_url=darks_url,
        raise_error_if_url_empty=False,
        save=True,
        overwrite=True,
    )
    copy_h5_dict_flats_to(
        scan=scan,
        flats_url=flats_url,
        raise_error_if_url_empty=False,
        save=True,
        overwrite=True,
    )
    assert scan.reduced_darks not in (None, {})
    assert scan.reduced_flats not in (None, {})
    assert numpy.array_equal(
        numpy.array(scan.reduced_darks_infos.count_time), numpy.array([1.0])
    )
    assert numpy.array_equal(
        numpy.array(scan.reduced_darks_infos.machine_current),
        numpy.array([12.3]),
    )
    assert numpy.array_equal(
        numpy.array(scan.reduced_flats_infos.count_time), numpy.array([2.0, 2.1])
    )
    assert numpy.array_equal(
        numpy.array(scan.reduced_flats_infos.machine_current),
        numpy.array([14.3, 14.2]),
    )


def test_reduced_frame_conversion(tmp_path):
    """
    test from_relative_reduced_frames_to_absolute and from_absolute_reduced_frames_to_relative functions
    """
    # ressources
    hdf5_scan_path = str(tmp_path / "my_hdf5")
    hdf5_scan = MockNXtomo(
        scan_path=hdf5_scan_path,
        dim=10,
        n_proj=100,
        n_ini_proj=100,
        create_ini_flat=False,
        create_ini_dark=False,
        create_final_flat=False,
    ).scan
    assert len(hdf5_scan.projections) == 100

    edf_scan_path = str(tmp_path / "edf_folder")
    MockEDF.mockScan(
        nRadio=20,
        scanID=edf_scan_path,
        n_extra_radio=2,
        scan_range=180,
        dim=36,
    )

    edf_scan = EDFTomoScan(edf_scan_path)
    assert edf_scan.tomo_n == 20

    # check type input
    with pytest.raises(TypeError):
        from_relative_reduced_frames_to_absolute(reduced_frames=1, scan=hdf5_scan)
    with pytest.raises(TypeError):
        from_relative_reduced_frames_to_absolute(reduced_frames={}, scan=1)
    with pytest.raises(TypeError):
        from_absolute_reduced_frames_to_relative(reduced_frames=1, scan=hdf5_scan)
    with pytest.raises(TypeError):
        from_absolute_reduced_frames_to_relative(reduced_frames={}, scan=1)

    # actual test
    assert from_absolute_reduced_frames_to_relative(
        reduced_frames={0: 0, 10: 1, 100: 2},
        scan=hdf5_scan,
    ) == {
        "0.0r": 0,
        "0.1r": 1,
        "1.0r": 2,
    }

    reduced_frames = {0: 0, 10: 1, 100: 2}
    assert from_relative_reduced_frames_to_absolute(
        from_absolute_reduced_frames_to_relative(
            reduced_frames=reduced_frames,
            scan=hdf5_scan,
        ),
        scan=hdf5_scan,
    )

    assert from_absolute_reduced_frames_to_relative(
        reduced_frames={0: 0, 10: 1, 30: 2},
        scan=edf_scan,
    ) == {
        "0.0r": 0,
        "0.5r": 1,
        "1.5r": 2,
    }

    # insure result won't changed if called twice
    assert from_absolute_reduced_frames_to_relative(
        reduced_frames=from_absolute_reduced_frames_to_relative(
            reduced_frames={0: 0, 10: 1, 30: 2},
            scan=edf_scan,
        ),
        scan=edf_scan,
    ) == {
        "0.0r": 0,
        "0.5r": 1,
        "1.5r": 2,
    }

    assert from_relative_reduced_frames_to_absolute(
        reduced_frames={0: 0, 10: 1, 30: 2},
        scan=edf_scan,
    ) == {0: 0, 10: 1, 30: 2}


__rlimit = resource.getrlimit(resource.RLIMIT_NOFILE)[0]


@pytest.mark.skipif(__rlimit > 1024, reason="rlimit > 1024")
def test_check_possible_issue_with_rlimit(tmp_path):
    tmp_dir = tmp_path / "test_check_possible_issue_with_rlimit"
    tmp_dir.mkdir()

    with NXtomoMockContext(
        scan_path=str(tmp_dir),
        n_proj=10,
        n_ini_proj=10,
        create_ini_dark=True,
        create_ini_ref=True,
        create_final_ref=False,
    ) as scan:
        # first test if no vds
        detector_url = DataUrl(
            file_path=scan.master_file,
            data_path=scan.get_detector_data_path(),
        )
        check_possible_issue_with_rlimit(
            url=detector_url, raise_error=True, delta_n_file=1024
        )

        # modify 'data' dataset to set a virtual dataset with broken link (file does not exists)
        with h5py.File(scan.master_file, mode="a") as h5f:
            detector_grp = h5f[scan.entry]["instrument/detector"]
            shape = detector_grp["data"].shape
            del detector_grp["data"]

            # create invalid VDS
            layout = h5py.VirtualLayout(shape=shape, dtype="i4")

            # this is a false 'vds' - only existances of files is checked
            filename = "toto.h5"
            vsource = h5py.VirtualSource(filename, "data", shape=shape)
            layout[0 : shape[0]] = vsource

            detector_grp.create_virtual_dataset("data", layout)

        # second test with vds
        check_possible_issue_with_rlimit(
            url=detector_url, raise_error=True, delta_n_file=0
        )

        check_possible_issue_with_rlimit(url=detector_url, delta_n_file=0)

        if __rlimit < 1000:
            check_possible_issue_with_rlimit(
                url=detector_url, raise_error=True, delta_n_file=1024
            )


def test_series_utils():
    """test `get_n_series` and `get_series_slice`"""
    images_keys = (
        ImageKey.DARK_FIELD.value,
        ImageKey.DARK_FIELD.value,
        ImageKey.FLAT_FIELD.value,
        ImageKey.FLAT_FIELD.value,
        ImageKey.FLAT_FIELD.value,
        ImageKey.PROJECTION.value,
        ImageKey.PROJECTION.value,
        ImageKey.PROJECTION.value,
        ImageKey.PROJECTION.value,
        ImageKey.FLAT_FIELD.value,
        ImageKey.FLAT_FIELD.value,
        ImageKey.FLAT_FIELD.value,
    )
    # test get_n_series
    assert (
        get_n_series(image_key_values=images_keys, image_key_type=ImageKey.PROJECTION)
        == 1
    )

    assert (
        get_n_series(image_key_values=images_keys, image_key_type=ImageKey.ALIGNMENT)
        == 0
    )

    assert (
        get_n_series(image_key_values=images_keys, image_key_type=ImageKey.DARK_FIELD)
        == 1
    )
    assert (
        get_n_series(image_key_values=images_keys, image_key_type=ImageKey.FLAT_FIELD)
        == 2
    )

    # test get_series_slice
    assert get_series_slice(
        image_key_values=images_keys, image_key_type=ImageKey.DARK_FIELD, series_index=0
    ) == slice(0, 2, 1)
    assert get_series_slice(
        image_key_values=images_keys,
        image_key_type=ImageKey.DARK_FIELD,
        series_index=-1,
    ) == slice(0, 2, 1)
    assert (
        get_series_slice(
            image_key_values=images_keys,
            image_key_type=ImageKey.DARK_FIELD,
            series_index=1,
        )
        is None
    )
    assert get_series_slice(
        image_key_values=images_keys, image_key_type=ImageKey.FLAT_FIELD, series_index=0
    ) == slice(2, 5, 1)
    assert get_series_slice(
        image_key_values=images_keys, image_key_type=ImageKey.FLAT_FIELD, series_index=1
    ) == slice(9, 12, 1)
    assert get_series_slice(
        image_key_values=images_keys, image_key_type=ImageKey.PROJECTION, series_index=0
    ) == slice(5, 9, 1)
    assert (
        get_series_slice(
            image_key_values=images_keys,
            image_key_type=ImageKey.ALIGNMENT,
            series_index=0,
        )
        is None
    )

    # test if images keys are empty
    assert (
        get_series_slice(
            image_key_values=(), image_key_type=ImageKey.DARK_FIELD, series_index=0
        )
        is None
    )
    assert (
        get_series_slice(
            image_key_values=(), image_key_type=ImageKey.DARK_FIELD, series_index=-1
        )
        is None
    )


def test_get_parameters_frm_par_or_info(tmp_path):
    """test the get_parameters_frm_par_or_info function"""
    info_file = os.path.join(tmp_path, "test.info")
    with open(info_file, mode="w") as f:
        f.writelines(
            [
                "Prefix=                 P40_highenergy_100nm_v3_1_\n",
                "Directory=              /data/visitor/blc15659/id16b/20240913/RAW_DATA/Vlad/P40_highenergy_100nm_v3_1_\n",
                "Energy=                 29.4\n",
                "Distance=               91.713\n",
                "ScanRange=              360\n",
                "TOMO_N=                 3203\n",
                "REF_ON=                 3204\n",
                "REF_N=                  21\n",
                "DARK_N=                 20\n",
                "Y_STEP=                 1.0\n",
                "Z_STEP=                 0\n",
                "Dim_1=                  2048\n",
                "Dim_2=                  2048\n",
                "Count_time=             0.07\n",
                "Latency_time (s)=       0.002\n",
                "Shutter_time=           7\n",
                "Col_end=                2047\n",
                "Col_beg=                0\n",
                "Row_end=                2047\n",
                "Row_beg=                0\n",
                "Optic_used=             0.65\n",
                "PixelSize=              0.0999\n",
                "Date=                   Fri Sep 13 20:24:06 2024\n",
                "Scan_Type=              continuous\n",
                "CCD_Mode=               FFM\n",
                "SrCurrent=              196.738\n",
                "CTAngle=                90\n",
                "Comment=                \n",
            ]
        )

    params = get_parameters_frm_par_or_info(info_file)
    assert params["energy"] == 29.4


def test_from_bliss_original_file_to_raw():
    """warning from_bliss_original_file_to_raw return the parent folder of the bliss original file"""
    assert (
        from_bliss_original_file_to_raw("/mnt/multipath-shares/data/toto/my_file.hdf5")
        == "/data/toto"
    )
    assert (
        from_bliss_original_file_to_raw("/gpfs/easy/data/toto/my_file.hdf5")
        == "/data/toto"
    )
    assert (
        from_bliss_original_file_to_raw("/gpfs/jazzy/data/toto/my_file.hdf5")
        == "/data/toto"
    )
    assert from_bliss_original_file_to_raw("/data/toto/my_file.hdf5") == "/data/toto"
    assert (
        from_bliss_original_file_to_raw("/home/data/toto/my_file.hdf5")
        == "/home/data/toto"
    )
    assert (
        from_bliss_original_file_to_raw("/gpfs/easy/home/data/toto/my_file.hdf5")
        == "/gpfs/easy/home/data/toto"
    )
    assert (
        from_bliss_original_file_to_raw("/home/user/documents/data/toto/my_file.hdf5")
        == "/home/user/documents/data/toto"
    )



import pytest
from tomoscan.esrf.scan.fscan_scantypes import scan_classes
from tomoscan.tests.datasets import GitlabDataset


@pytest.fixture(scope="class")
def bootstrap_test_scantypes(request):
    cls = request.cls
    cls.scan_fname = {}
    for scan_type in ["aeroystepscan", "fscan", "limatake", "loopscan"]:
        cls.scan_fname[scan_type] = GitlabDataset.get_dataset(
            f"h5_datasets/fscan/{scan_type}.h5"
        )

    yield
    # Tear-down


@pytest.mark.usefixtures("bootstrap_test_scantypes")
class TestScanTypes:
    """
    Test scan types for the different type of 'scan' of FScan
    """

    def test_scantype(self):
        for scan_type, scan_fname in self.scan_fname.items():
            scan_cls = scan_classes[scan_type]
            scan = scan_cls(scan_fname)
            metadata = scan.get_metadata()
            expected_fields = set(scan_cls.required_fields)

            # "metadata" might have more items (see 'optional_fields').
            # The following difference is the empty set iff 'metadata' has at least the items of 'expected_fields'
            diff = expected_fields - set(metadata.keys())
            diff_without_names_templates = set(item for item in diff if "{" not in item)
            assert (
                len(diff_without_names_templates) == 0
            ), "Expected fields %s for scan type %s, but got %s" % (
                str(expected_fields),
                scan_type,
                str(set(metadata.keys())),
            )



# coding: utf-8

import pytest

from tomoscan.esrf.scan.mock import MockNXtomo
from tomoscan.framereducer.framereducerbase import FrameReducerBase


def test_FrameReducerBase_instanciation(tmp_path):
    scan = MockNXtomo(tmp_path, n_proj=2).scan
    reducer = FrameReducerBase(scan=scan, reduced_method="mean", target="darks")
    with pytest.raises(NotImplementedError):
        reducer.run()



# coding: utf-8
"""module to mock volume"""

from __future__ import annotations

from typing import Sized

import numpy
from silx.image.phantomgenerator import PhantomGenerator
from enum import Enum

__all__ = ["Scene", "create_volume"]


class Scene(Enum):
    SHEPP_LOGAN = "Shepp-Logan"


def create_volume(
    frame_dims: int | tuple, z_size: int, scene: Scene = Scene.SHEPP_LOGAN
) -> numpy.ndarray:
    """
    create a numpy array of the requested scheme for a total of frames_dimes*z_size elements

    :param tuple frame_dims: 2d tuple of frame dimensions
    :param z_size: number of elements on the volume on z axis
    :param Scene scene: scene to compose
    """
    scene = Scene(scene)
    if not isinstance(z_size, int):
        raise TypeError(
            f"z_size is expected to be an instance of int not {type(z_size)}"
        )
    if scene is Scene.SHEPP_LOGAN:
        if isinstance(frame_dims, Sized):
            if not len(frame_dims) == 2:
                raise ValueError(
                    f"frame_dims is expected to be an integer or a list of two integers. Not {frame_dims}"
                )
            if frame_dims[0] != frame_dims[1]:
                raise ValueError(
                    f"{scene} only handle square frame. Frame width and height should be the same"
                )
            else:
                dim = frame_dims[0]
        elif isinstance(frame_dims, int):
            dim = frame_dims
        else:
            raise TypeError(
                f"frame_dims is expected to be a list of two integers or an integer. Not {frame_dims}"
            )
        return numpy.asarray(
            [PhantomGenerator.get2DPhantomSheppLogan(dim) * 10000.0] * z_size
        )
    else:
        raise NotImplementedError
